import fastlmm.association.lrt as lr
import fastlmm.association.score as score
import fastlmm.association.testCV as testCV
import fastlmm.inference as fastlmm
from fastlmm.pyplink.plink import *
from pysnptools.util.pheno import *
from fastlmm.pyplink.altset_list import *
import subprocess, sys, os.path
import fastlmm.util.stats.chi2mixture as c2
from fastlmm.util.distributable import *
from fastlmm.util.runner import *
import pdb
import fastlmm.util.util as utilx
import fastlmm.util.stats as ss
import time
from Result import *
from PairResult import *
from fastlmm.association.tests import *
import fastlmm.util.genphen as gp
import scipy as sp
from itertools import *
from fastlmm.pyplink.snpreader.Bed import *
import logging
import warnings
from tempfile import TemporaryFile
import fastlmm.util.preprocess as util

class FastLmmSet: # implements IDistributable
    '''
    A class for specifying a FastLmmSet and then running it.
    '''


    def addpostifx_to_outfile(self):
        if self.datestamp is not None:
            if self.datestamp=="auto":
                self.datestamp=utilx.datestamp()
            self.outfile=utilx.appendtofilename(self.outfile,self.datestamp,"_")

    def __init__(self, **entries):
        '''
        outfile         : string of the filename to which to write out results (two files will be generated from this, on a record of the run)
        phenofile       : string of the filename containing phenotype
        alt_snpreader   : A snpreader for the SNPs for alternative kernel. If just a file name is given, the Bed reader (with standardization) is used.
        bedfilealt      : (deprecated) same as alt_snpreader
        filenull        : string of the filename (with .bed or .ped suffix) containing SNPs for null kernel. Should be Bed format if autoselect is true.
        extractSim      : string of a filename containing a list of whitespace-delmited SNPs. Only these SNPs will be used when filenull is read.
                               By default all SNPs in filenull are used.
                               It is an error to specify extractSim and not filenull.
                               Currently assumes that filenull is in bed format and that autoselect is not being used.
        altset_list     : list of the altsets
                               By default this is a file that will be read by SnpAndSetNameCollection
                               but a file of nuc ranges can be given again via 'NucRangeSet(filename)'
        altset_list2    : if this is present, then does an interaction test between genes in altset_list and alt_setlist2
        covarfile       : string of the filename containing covariate
        mpheno=1        : integer representing the index of the testing phenotype (starting at 1)
        ipheno=0        : same as mpheno, but starts at 0 (deprecated)
        mindist=2.0     : SNPs within mindist from the alternative SNPs will be removed from
                          the null kernel computation
        idist=1         : the index of the position index to use with mindist
                            1 : genomic distance
                            2 : base-pair distance
        autoselect = True   : Should autoselect to used?
        test="lrt"      :['sc_davies', 'sc_mom', 'sc_liuskat','sc_daviesskat','sc_all','lrt','lrt_up']
        nperm = 10       : number of pemutations per test
        npermabs = None : absolutely number of permutations
        calseed =   : the int that gets added to the random seed 34343 used to generate permutations for lrt null fitting.
        write_lrtperm   : If true, write the lrtperm vector (dictated by calseed) to a file.
        forcefullrank   :  If true, the covariance is always computed for lrt (default False)
        minsetsize      : If non-None, only include sets at least this large (inclusive)
        maxsetsize      : If non-None, only include sets no more than this large (inclusive)
        datestamp       : Defaults to None. Use "auto" to have it generate a unique date and time string
                          to append to filenames. Otherwise, it appends whatever you set it to.
        log             : (Defaults to not changing logging level) Level of log messages, e.g. logging.CRITICAL, logging.ERROR, logging.WARNING, logging.INFO
        '''

        # member variables with default values, or those not defined in the input file    
        self.mpheno=1
        self.mindist = 2.0
        self.idist=1
        self.autoselect = True
        self.extractSim = None
        self.nperm = 10 # used to permute the data in order to get empirical distrib of null for param fitting of null; these are global
        self.npermabs = None #used instead of nperm to compute an at least this many permutations (I say at least because it is # of perms per test)
        self.cache_from_perm=False #If true, then actually does the permutations with the real, all in run_test() so can cache expensive things
        self.nlocalperm=None #same as nperm, but local to each test (doesn't do pooling)
        self.fitlocal=False #if true, fit aUD to each local set
        self.write_lrtperm = False  #when set to True, writes a third file containing only the p-values resulting from nperm
        self.nullfitfile = None   #file from which to grab the null stats from. should be generated using write_lrtperm
        self.calseed = None# was 0 # the int that gets added to the random seed 34343 used to generate permutations for lrt null fitting
        self.sets=None
        self.permute=None       # permute the SNPs being tested (or, equivalently, the compliment) using this with 1,2,3,4, etc. for different seeds, used for type I error tables
        self.forcefullrank = False
        self.nullfit="qq"       #alternative is "ml"    
        self.qmax=0.1
        self.minsetsize = 1
        self.maxsetsize = 10000000
        self.datestamp = None #use auto to have it auto-generate a date stamp
        self.covarimp = 'standardize' #use 'standardize' for mean and variance standardization, with mean imputation
        self.altset_list2=None#for doing pairs of sets
        self.nullModel = None
        self.altModel = None
        self.genphen=None#MUST LEAVE THIS AS THE DEFAULT#{"varE":1.0,"varG":1.0, "varBack":1.0,"varCov":1.0,"link":'linear',"casefrac":0.5,"once":false}      #generate synthetic phen, using LMM, and real SNPs
       
        self.scoring = None
        self.greater_is_better = None
        self.log = None
        self.detailed_table = False
        self.signal_ratio = True

        self._synthphenfile=None

        self.alt_snpreader = None
        self.show_pvalue_5050 = False

        
        if hasattr(self,"ipheno"):
            assert not hasattr(self,"mpheno")
            self.mpheno = self.ipheno+1

        self.__dict__.update(entries)
        self._check_entries(entries)           
        self._ran_once = False
        self.addpostifx_to_outfile()

        if self.hasNonNoneAttr("alt_snpreader") and self.hasNonNoneAttr("bedfilealt") : raise Exception("alt_snpreader or bedfilealt may be given, but not both")
        if self.hasNonNoneAttr("bedfilealt"):
            warnings.warn("bedfilealt is deprecated. alt_snpreader instead", DeprecationWarning)
            self.alt_snpreader = self.bedfilealt
            delattr(self,"bedfilealt")
        if isinstance(self.alt_snpreader, str):
            self.alt_snpreader = Bed(self.alt_snpreader)


        # convert deprecated "verbose" into "log"        
        if self.hasNonNoneAttr("log") and self.hasNonNoneAttr("verbose") :
            raise Exception("log or verbose may be given, but not both")
        if self.hasNonNoneAttr("verbose"):
            if self.verbose:
                self.log = logging.INFO
            else:
                self.log = logging.CRITICAL
            delattr(self, "verbose")

        if self.hasNonNoneAttr("log"): #If neither were set
            logger = logging.getLogger()
            logger.setLevel(self.log)


        if not hasattr(self, "test") : raise Exception("FastLmmSet must have 'test' set")
        #allowed_tests=['sc_davies', 'sc_mom', 'sc_liuskat','sc_daviesskat','sc_all','lrt','cv_l1','cv_l2']
        #if self.test not in allowed_tests: raise Exception("test must be one of " + str(allowed_tests) +", but found " + self.test)

    # =================================================================
    # BEGIN Implementation of the IDistributable interface described in fastlmm.util.distributable
    # =================================================================

    def hasNonNoneAttr(self, attr):
        return hasattr(self,attr) and getattr(self,attr) != None


    # required by IDistributable
    @property
    def work_count(self):
        '''
        Tells how many work items there are
        '''
        self.run_once() #load files, etc. -- stuff we only want to do once per task (e.g. on the cluster)
                        #no matter how many times we call 'run_test' (and then 'reduce' which of course
                        #only gets run one time, and calls this too, but doesn't actually do the work).

        if self.altset_list2 is None: #singleton sets
            return len(self.altsetlist_filtbysnps) * (1+self.nperm)
        else: #pairs of sets
            return len(self.altsetlist_filtbysnps)*len(self.__altsetlist2_filtbysnps)* (1+self.nperm)   

    def printPhenToFile(self, nInd, SNPsalt, y):
        self._synthphenfile=utilx.appendtofilename(self.outfile,"synthPhenVarBack")
        np.savetxt(self._synthphenfile,y)
        with open(self._synthphenfile,"w") as fp: 
            for n in sp.arange(0,nInd):
                 fp.write(SNPsalt['iid'][n][0] +'\t'+ SNPsalt['iid'][n][1]+ "\t" + '%f' % y[n] + "\n")

    def work_sequence(self):
        '''
        Enumerates a sequence of work items
        Each work item is a lambda expression (i.e. function pointer) that calls 'run_test', returning a list of Results (often just one)
        '''
        self.run_once() #load files, etc. -- stuff we only want to do once per task (e.g. on the cluster)
                        #no matter how many times we call 'run_test' (and then 'reduce' which of course
                        #only gets run one time, and calls this too, but doesn't actually do the work).
        ttt0=time.time()
        y=None        
        y_back=None
        nInd=len(self.__y)               
        haswrittenphen=False;                            
                
        if self.altset_list2 is None: #singleton sets            
            for iset, altset in enumerate(self.altsetlist_filtbysnps):
                for iperm in xrange(-1, self.nperm):   #note that self.nperm is the 'stop', not the 'count'
                    SNPsalt=altset.read()
                    SNPsalt['snps'] = util.standardize(SNPsalt['snps'])
                    G1 = SNPsalt['snps']/sp.sqrt(SNPsalt['snps'].shape[1])  
                    ichrm =  ",".join(sp.array(sp.unique(SNPsalt['pos'][:,0]),dtype=str)) 
                    minpos= str(sp.min(SNPsalt['pos'][:,2]))
                    maxpos= str(sp.max(SNPsalt['pos'][:,2]))
                    iposrange = minpos + "-" + maxpos                     
                    
                    if self.genphen is None:
                        y=self.__y
                    else: 
                        assert self.permute is None, "Error: using permute with genphen--there should be no need, use genphen['seed'] instead"
                        self.__y=None      

                        #if (self.__G0 is not None) and y_G0 is None:  #cache so only do it once
                        if (self.__SNPs0 is not None) and y_G0 is None:  #cache so only do it once   
                            newseed = self.mainseed ^ self.genphen['seed']
                            from numpy.random import RandomState
                            randomstate = RandomState(newseed)
                            nSnp=self.__SNPs0['data']['snps'].shape[1]
                            #good for low rank, other wise, use the dual, as in gp.genphen
                            raise Exception("Bug below. should be randn, is rand")
                            y_G0=sp.sqrt(self.genphen["varBack"]/nSnp)*self.__SNPs0['data']['snps'].dot(randomstate.rand(nSnp,1))    #TODO: CL This seems to be a bug. Should be randn
                            #y_G0=sp.sqrt(self.genphen["varBack"])*self.__SNPs0['data']['snps'].dot(randomstate.rand(nSnp,1))    #TODO: CL This seems to be a bug. Should be randn
                        elif self.__SNPs0 is None:
                            y_G0=0      

                        #always have the same background signal                                                
                        if self.genphen["varBackNullFileGen"] is not None and y_back is None:                                                      
                            nSnp=self.__SNPs0['data']['snps'].shape[1]
                            y_back=sp.sqrt(self.genphen["varBack"]/nSnp)*self.__varBackNullSnpsGen['snps'].dot(sp.random.randn(self.__varBackNullSnpsGen['snps'].shape[1],1))/sp.sqrt(self.__varBackNullSnpsGen['snps'].shape[1]) 
                            #self.printPhenToFile(nInd, SNPsalt, y_back);
                        elif self.genphen.has_key("varBackNullPhenGen") and self.genphen["varBackNullPhenGen"] is not None and y_back is None:
                            y_back=loadPhen(filename = self.genphen["varBackNullPhenGen"])['vals']                            
                        elif y_back is None:          
                            y_back=0                          
                        
                        if self.genphen["once"] and y is None: # only generate the phenotype once for entire run                            
                            newseed = self.mainseed ^ self.genphen['seed']  
                            assert  self.genphen["varG"]==0, "doesn't make sense to have varG>0 and only one phen--not sure why I put this code there"       
                            if self.genphen["varG"]>0 and nInd<=SNPsalt['snps'].shape[1]:                                                       
                                Kall = self.KfromAltSnps(nInd) #useful for full rank
                                y=gp.genphen(y_G0=y_G0+y_back,G1=None,covDat=self.__X,options=self.genphen,nInd=nInd,
                                             K1=Kall,randseed=newseed)                          
                            else:     
                                Kall=None                                     
                                y=gp.genphen(y_G0=y_G0+y_back,G1=G1,covDat=self.__X,options=self.genphen,nInd=nInd,randseed=newseed)
                        elif (not self.genphen["once"]):  #generate for each set in turn                           
                            eachseed=utilx.combineseeds(iset,self.genphen['seed'])                            
                            newseed = self.mainseed ^ eachseed                              
                            y=gp.genphen(y_G0=y_G0+y_back,G1=G1,covDat=self.__X,options=self.genphen,nInd=nInd,randseed=newseed) 
                                                                
                    assert y is not None, "y is None"                                   
                    yield lambda altset=altset,iset=iset,iperm=iperm,y=y,ichrm=ichrm, iposrange=iposrange : self.run_test(SNPs1=SNPsalt,G1=G1, y=y, altset=altset, iset=iset, iperm=iperm, ichrm=ichrm, iposrange=iposrange)
        else: #pairs of sets
            raise Exception("not implemented, started a long time ago and never finished")
            #for iperm in xrange(-1, self.nperm):   #note that self.nperm is the 'stop', not the 'count'
            #    for iset, altset in enumerate(self.__altsetlist_filtbysnps):
            #         for iset2, altset2 in enumerate(self.__altsetlist2_filtbysnps):
            #            if iset!=iset2:
            #                yield lambda altset=altset,iset=iset,altset2=altset2,iset2=iset2,iperm=iperm : self.run_interactiontest(altset,iset,altset2,iset2,iperm)

        ttt1=time.time()
        logging.info("---------------------------------------------------")
        logging.info("Elapsed time for all tests is %.2f seconds" % (ttt1-ttt0))
        logging.info("---------------------------------------------------")
    
    def check_for_None(self, alteqnull, alteqnullperm):
        none_ind = []
        for i in range(len(alteqnull)):
            if alteqnull[i] is None:
                none_ind.append(i)
        none_ind_perm = []
        for i in range(len(alteqnullperm)):
            if alteqnullperm[i] is None:
                none_ind_perm.append(i)
        if len(none_ind)>0:
            raise Exception("found None entries in alteqnull during reduce, which means some Result objects were missing")
        if len(none_ind_perm)>0:
            raise Exception("found None entries in alteqnullperm during reduce, which means some Result objects were missing")

    # required by IDistributable
    def reduce(self, result_list_sequence):
        '''
        Given a sequence of results from 'run_test', create the output report.

        '''
        logging.info("last modified: %s" % time.ctime(os.path.getmtime(__file__)))
        logging.info("-------------------------------------------------")

        # Create the info file and direct all 'stdout' messages to both stdout and to the info file
        infofile=utilx.appendtofilename(self.outfile,"info")
        outfiletab=self.outfile
        if self.datestamp=="auto": self.outfile=utilx.appendtofilename(self.outfile,"tab")

        try:
            utilx.create_directory_if_necessary(infofile)
        except:
            logging.warn("Exception while creating directory for '{0}'. Assuming that other cluster task is creating it.".format(infofile))

        logging_handler=logging.FileHandler(infofile,"w",delay=False)
        logger = logging.getLogger()
        logger.addHandler(logging_handler)

        logging.info("distributable = " + self.__repr__())

        self.run_once() #load files, etc. -- doesn't actually do the work if it's already been done
                         #note, however, that before reduce is called, work_count calls run_once() anyhow

        result_dict = {}

        lrt = SP.nan*SP.ones(len(self.altsetlist_filtbysnps))
        # whether alt and null models give the same margll
        alteqnull = [None]*len(self.altsetlist_filtbysnps)

        lrtperm = SP.nan*SP.ones(len(self.altsetlist_filtbysnps)*self.nperm)
        alteqnullperm = [None]*len(self.altsetlist_filtbysnps)*self.nperm
        setsizeperm = SP.nan*SP.ones(len(self.altsetlist_filtbysnps)*self.nperm)

        npvals=self.test.npvals
        pv_adj= sp.nan*sp.ones((len(self.altsetlist_filtbysnps)))
              
        # results can come in any order, so we have to use iperm and iset to put them in the right place
        # there is one result instance for each combination of test and permutation, and here we are just gathering them
        # into the arrays from above

        for result_list in result_list_sequence:
            for result in result_list:
                if result.iperm < 0:
                    result_dict[result.iset] = result
                    #iset is the index of the test (irrespective of permutation)
                    lrt[result.iset] = self.test.lrt(result)
                    alteqnull[result.iset] = result.alteqnull #equiv to result["alteqnull"]
                    #pv_adj[result.iset,:] = self.test.pv_adj_from_result(result)                
                    pv_adj[result.iset] = self.test.pv_adj_from_result(result)                
                else:
                    isetiperm = result.iset+result.iperm*len(self.altsetlist_filtbysnps)
                    lrtperm[isetiperm] = self.test.lrt(result)
                    alteqnullperm[isetiperm] = result.alteqnull
                    setsizeperm[isetiperm] = result.setsize
                if result.alteqnull is None and str(self.test)[0:3]=="lrt":
                    raise Exception("self.alteqnull is None")

        #look for None to see if anything didn't get filled in, and if so, where
        if str(self.test)[0:3]=="lrt":
            self.check_for_None(alteqnull, alteqnullperm)
            
        if self.write_lrtperm and self.nperm>0:
            self._saveArray("lrtperm", ["2*(LL(alt)-LL(null))","alteqnull","setsize"], (lrtperm,alteqnullperm,setsizeperm))
                
        pv_adj,ind = self.test.pv_adj_and_ind(self.nperm, pv_adj, self.nullfit, 
                                              lrt, lrtperm, alteqnull, alteqnullperm,self.qmax, self.nullfitfile, self.nlocalperm)

        logging.info("writing the result files : " + self.outfile + "")

        #TODO: move the following to a function/object:
        with open(outfiletab,"w") as fp:
            if type(self.test) is Lrt: #dan: this is ugly, will fix soon                
                self.test.write(fp, ind, result_dict, pv_adj, self.detailed_table, self.signal_ratio)
            else:
                self.test.write(fp, ind, result_dict, pv_adj, self.detailed_table)

        if not self.show_pvalue_5050: #Remove the unwanted column if requested and necessary
            dataframe=pd.read_csv(outfiletab,delimiter='\t',comment=None) #Need \t instead of \s because the output has tabs by design and spaces in column names(?)
            # adjust the dataframe and then remove the original outfile and (if requested) create a final outfile
            header5050 = 'P-value(50/50)'
            if header5050 in dataframe.columns:
                dataframe.drop(header5050, axis=1, inplace=True)
            dataframe.rename(columns={'P-value_adjusted': 'P-value'}, inplace=True)
            dataframe.to_csv(outfiletab, sep="\t", index=False)

        logger.removeHandler(logging_handler)
        logging_handler.close()
        return self.outfile
        
    def getRandSnpSignal(self, nSnp, nInd,genphen,newseed):
        from numpy.random import RandomState
        import fastlmm.util.gensnp as gp
        randomstate = RandomState(newseed) 
        nSnp=genphen["numBackSnps"]
        #randsnps=self.alt_snpreader.read(RandomSnpSet(nSnp,newseed))     #this appears to be VERY slow
        #snps=randsnps['snps']         
        snps=gp.gensnps(nInd,nSnp)                        
        randG = snps/sp.sqrt(nSnp) #pre-process for kernel
        y_randG=sp.sqrt(genphen["varBack"])*randG.dot(randomstate.rand(nSnp,1))            
        return y_randG

    def TESTBEFOREUSINGKfromAltSnps(self, N, SNPsalt=None):        
       
        print "constructing K from all SNPs"
        t0=time.time()

        Kall=sp.zeros([N,N])
        nSnpTotal=self.alt_snpreader.snp_count
        print "reading in " +str(nSnpTotal)+ " SNPs and adding up kernels"
        #altsnps = readBED(self.bedfilealt,standardizeSNPs=True)['snps'] #loads all in to memory        
        blocksize=100
        ct=0
        ts=time.time()
        if self.alt_snpreader.snp_count<N: raise Exception("need to adjust code to handle low rank")          
        for start in range(0,self.alt_snpreader.snp_count,blocksize):                                
            ct+=blocksize
            snpSet = PositionRange(start,blocksize)
            snps = self.alt_snpreader.read(snpSet)['snps']
            import fastlmm.util.preprocess as util
            snps = util.standardize(snps)

            #print "start = {0}".format(start)
            Kall=Kall + snps.dot(snps.T)    
            t1=time.time()
            if ct % 50000==0: 
                print "read %s SNPs in %.2f seconds" % (ct, t1-ts)
                ts=time.time()
            #if ct==2: 
            #    break
            #    Kall=sp.rand((N,N))
            #    Kall=Kall.dot(Kall.T)
        Kall=Kall/sp.sqrt(self.alt_snpreader.snp_count)  
        Kall=Kall + 1e-5*sp.eye(N,N)     
        t1=time.time()
        logging.info("%.2f seconds elapsed" % (t1-t0))        
                
        return Kall

    def _saveArray(self, appendNameFile, header, values):
        outfile = utilx.appendtofilename(self.outfile, appendNameFile)
        try:
            utilx.create_directory_if_necessary(outfile)
        except:
            logging.warn("Exception while creating directory for '{0}'. Assuming that other cluster task is creating it.".format(outfile))

        logging.info("writing to file " + outfile + ".")

        assert type(values) == sp.ndarray or type(values) == tuple, 'values can only be sp.ndarray or a tuple of sp.ndarray.'

        if type(values) == sp.ndarray:
            values = (values,)

        values = map( list, zip(*values) )

        with open(outfile, "w") as fp:
            fp.write('\t'.join(header) + "\n")
            for v in values:
                fp.write("{0}\n".format( '\t'.join([str(x) for x in v]) ))
    
    #!! would be nice of this was optional and if not given the OS was asked
    # required by IDistributable
    @property
    def tempdirectory(self):
        return self.outfile + ".work_directory"


    #!! need comments
    def copyinputs(self, copier):
        copier.input(self.phenofile)
        copier.input(self.alt_snpreader)
        copier.input(self.altset_list)
        if (self.altset_list2 is not None):
            copier.input(self.altset_list2)
        copier.input(self.covarfile)
        # similar code elsewhere
        if (self.filenull is not None):
            root, ext = os.path.splitext(self.filenull)
            # allow other formats?
            if ext.lower() == ".bed":
                copier.input(root + ".bed")
                copier.input(root + ".bim")
                copier.input(root + ".fam")
            else:
                copier.input(root + ".ped")
                copier.input(root + ".map")
        if (self.extractSim is not None):
            copier.input(self.extractSim)
        if (self.nullfitfile is not None):                        
            copier.input(self.nullfitfile)
        if self.genphen is not None and self.genphen.has_key("varBackNullFileGen"):
            copier.input(self.genphen["varBackNullFileGen"]+".fam")
            copier.input(self.genphen["varBackNullFileGen"]+".bim")
            copier.input(self.genphen["varBackNullFileGen"]+".bed")
        if self.genphen is not None and self.genphen.has_key("varBackNullPhenGen") and self.genphen["varBackNullPhenGen"] is not None:
            copier.input(self.genphen["varBackNullPhenGen"])

    #Note that the files created are not automatically copied. Instead,
    # whenever we want another file to be created, a second change must be made here so that it will be copied.
    def copyoutputs(self,copier):
        outfiletab=self.outfile
        copier.output(outfiletab)
        infofilename = utilx.appendtofilename(self.outfile,"info")
        copier.output(infofilename)
        if self.write_lrtperm and self.nperm>0:
            lrtperm_outfile = utilx.appendtofilename(self.outfile,"lrtperm")
            copier.output(lrtperm_outfile)
        if self._synthphenfile is not None:
            copier.output(self._synthphenfile)

    # =================================================================
    # END Implementation the IDistributable interface described in fastlmm.util.distributable
    # =================================================================

    def _check_entries(self, entries):
        assert entries.has_key('nullModel') and entries.has_key('altModel')
        assert entries['nullModel'].has_key('effect') and entries['nullModel'].has_key('link')
        assert entries['altModel'].has_key('effect') and entries['altModel'].has_key('link')

        assert entries['nullModel']['effect'] in set(['fixed', 'mixed'])
        assert entries['altModel']['effect'] in set(['fixed', 'mixed'])

        assert entries['nullModel']['link'] in set(['linear', 'logistic', 'erf'])
        assert entries['altModel']['link'] in set(['linear', 'logistic', 'erf'])

        assert all( [(v in set(['effect', 'link', 'penalty', 'approx'])) for v in entries['nullModel']] ),\
               'The only allow keys for nullModel are effect, link, penalty, and approx.'

        assert all( [(v in set(['effect', 'link', 'penalty', 'approx'])) for v in entries['altModel']] ),\
               'The only allow keys for altModel are effect, link, penalty, and approx.'

       

    def _check_params_after(self):        
        self.nperm = self.test.check_nperm(self.nperm)
        if self.minsetsize is not None and self.minsetsize<1: raise Exception("minsetsize must be 1 or greater")
        if self.nullfitfile is not None: assert (self.nperm is None or self.nperm==0), "if using nullfitfile then nperm should be 0"
        assert ((self.nlocalperm is None) or self.nlocalperm==0) or ((self.nperm is None) or self.nperm==0), "cannot use both nperm and nlocalperm"
        assert self.nullfitfile is None or self.nlocalperm==0, "cannot use nullfitfile with nlocalperm"

        if self.genphen is not None:
            if 'casefrac' not in self.genphen.keys(): self.genphen['casefrac']=None
            if 'fracCausal' not in self.genphen.keys(): self.genphen['fracCausal']=1.0
            if 'numBackSnps' not in self.genphen.keys(): self.genphen['numBackSnps']=0
            if 'varBackNullFileGen' not in self.genphen.keys(): self.genphen['varBackNullFileGen']=None  
    
    def _check_params_before(self):
        pass
        #if self.test== "lrt_up" and self.filenull is None:
        #    self.test = "lrt"
                   
    def pv_from_localperm(self, iperm, result, t0, varcomp_test, y):
        '''
        This is not test specific, but is currently only implemented for LRT. No code would 
        need to be changed here for other tests, only the test-specific files elsewhere.
        '''
        assert iperm==-1, "should not be permuting for global null fit here, only local"
        assert str(self.test)=="lrt", "can only use local permutations with lrt, not with " + str(self.test)            
        assert self.altModel["effect"]=="mixed" and self.altModel["link"]=="linear", "nlocalperm only set up for linear lrt at the moment"
        assert self.nullModel["effect"]=="fixed", "nlocalperm only set up for one-kernel" 
        #assert self.__G0 is None, "only 1-kernel implemented"
        assert self.__SNPs0 is None, "only 1-kernel implemented"
        logging.info("running up to " + str(self.nlocalperm) + " local permutations")
        permstatbetter=[]
        allstat=[]
        alteqnullperm=[]
        yorig=y
        tol=0.0
        from numpy.random import RandomState
        randomstate = RandomState(self.rseed)
        checkpoint=1000
        powerneeded=2
        for pm in xrange(self.nlocalperm):
            if pm==checkpoint:
                logging.info('checkpointing '  + str(pm))                    
                numbetter=sp.sum(permstatbetter)
                pv=numbetter/float(pm)                                
                t1=time.time()
                logging.info("Checkpoint time elapsed=%.2f seconds" % (t1-t0))
                if pv*checkpoint>=10**powerneeded:
                    logging.info("stopping due to checkpoint")
                    break;
                checkpoint=checkpoint*2
            
            permutationIndex = utilx.generatePermutation(y.shape[0],randomstate)
            yperm=y[permutationIndex]           
            Xperm=self.__X[permutationIndex]                
                                                                   
            permresult = varcomp_test.testGupdate(yperm,Xperm,self.test)  
            allstat.append(permresult['stat'])
            alteqnullperm.append(permresult['alteqnull'])
            permstatbetter.append(permresult['stat']-result.test['stat']>tol)  
                       
        numbetter=sp.maximum(0.99,sp.sum(permstatbetter))
        pv=numbetter/float(pm) 
        result.test['pv-local']=pv
        result.test['pv']=result.test['pv']
        
        if self.fitlocal and pv*pm<10**powerneeded:
            import fastlmm.association.tests.Cv as cv
            logging.info("fitting aUD to local with " + str(pm+1) + " permutations")
            pv_adj,mixture,scale,dof=cv.lrtpvals_qqfit(nperm=self.nlocalperm, lrt=sp.array([result.test['stat']]), lrtperm=sp.array(allstat), alteqnull=sp.array([result.test['alteqnull']]), alteqnullperm=sp.array(alteqnullperm), qmax=self.qmax)
            result.test['pv-local-aUD']=pv_adj[0]
            logging.info("mixture (non-zero dof)="+str(mixture))
            logging.info("dof="+str(dof))
            logging.info("scale="+str(scale))
        else:
            result.test['pv-local-aUD']=sp.NaN
        
        logging.info("    used " + str(pm+1) + " permutations to compute p=" + str(pv) + ", p50=" + str(result.test['pv']))
        
    def G_exclude(self, i_exclude):
        if self.__SNPs0.has_key("data"):
            G_exclude = self.__SNPs0["data"]["snps"][:,i_exclude]
        else:
            snp_names = self.__SNPs0["reader"].rs[i_exclude]
            snp_set = SnpAndSetName('G_exclude', snp_names)
            G_exclude = self.__SNPs0["reader"].read(snp_set)['snps']
            G_exclude = util.standardize(G_exclude)
            #normalize
            pass
        G_exclude/=sp.sqrt(self.__SNPs0["num_snps"])
        return G_exclude

    def run_test(self, SNPs1, G1, y, altset, iset, ichrm, iposrange, iperm = -1, varcomp_test=None):
        '''
        This function does the main work of the class, and also reads in the SNPs for the alternative model.
        It is called (via a lambda) inside the loops found in 'generate_sequence'.

        Input:
            altset - a set of snps
            iset - index to altset
            iperm - index to permutation (-1 means no permutation)
            varcomp -if not None, assume that it is the correct one for this test, and does not re-compute anything
        Output:
            a list (often just one) of
                instances of the Result class, varcomp (for caching)
        '''
        t0=time.time()            

        self.run_once() #load files, etc. -- stuff we only want to do once no matter how many times we call 'run_test' and then 'reduce.        

        logging.info("Working on permutation index #{0} (where -1 means no permutation) out of {1} total permutations...".format(iperm, self.nperm))
        logging.info("\taltset {0}, {1} of {2}  ".format(altset, iset, len(self.altsetlist_filtbysnps)))
                      
        result = Result(iperm=iperm,iset=iset,setname=str(altset), ichrm=ichrm, iposrange=iposrange)
         
        if varcomp_test is None or not self.cache_from_perm:            
            if G1.shape[1]==0:
                logging.info( "no SNPS in set " + setname )
                result=None
                return [result]
            if sp.isnan(G1.sum()): raise Exception("found missing values in test SNPs that remain after intersection for " + str(altset))

            result.setsize = SNPs1['snps'].shape[1]
                
            logging.info(" (" + str(result.setsize) + " SNPs)")              

            if self.filenull is not None:
                if self.__SNPs0.has_key('data'):
                    pos0=self.__SNPs0['data']['pos']
                else:
                    pos0=self.__SNPs0['snp_set'].pos
                i_exclude =  utilx.excludeinds(pos0, SNPs1['pos'], mindist = self.mindist,idist = self.idist)
                result.nexclude = i_exclude.sum()
                logging.info("numExcluded=" + str(result.nexclude) + "")
            else:
                i_exclude = None
                result.nexclude=0;


        if self.permute is not None: #permute the data to test for type I error
                permutationIndex = utilx.generatePermutation(SNPs1['snps'].shape[0],self.mainseed ^ self.permute)      
        if iperm >= 0 :  #permute the data to create Null-only P-values for lrt null fitting
            eachseed=utilx.combineseeds(self.calseed,iperm)            
            newseed = self.mainseed ^ eachseed  

            if self.permute is not None :
                #Add a left shift so that iperm=1 and self.permute=2 gives a different newseed than 
                #iperm=2 and self.permute=1
                newseed = (newseed << 1) ^ self.permute
            permutationIndex = utilx.generatePermutation(SNPs1['snps'].shape[0], newseed)
        if varcomp_test is None or not self.cache_from_perm:
            #need to make this caching smarter for when background kernel changes in every test (e.g. interactions)
            if (self.permute is not None) or (iperm >=0):
                G1=G1[permutationIndex]

            #if result.nexclude:
            #    logging.info(" (computing needed null info anew) ")
            #    G0_to_use = self.__SNPs0#['data']['snps'][:,~i_exclude]/SP.sqrt(self.__SNPs0['data']['snps'][:,~i_exclude].shape[1])   
            #else: 
            #    G0_to_use=self.__G0
        
            if (str(self.test)!="lrt_up") and (self.genphen is not None) and (not self.genphen['once']) or ((str(self.test)!="lrt_up") and result.nexclude>0) or self.__varcomp_test is None:
                #need to re-construct the test (can't cache)
                #note that if result.nexclude kicked in, then this would happen in clause above anyhow                               
                varcomp_test = self.varcomp_test_setup(y,self.__SNPs0,i_exclude)
            else: #use cached values                
                #G0_to_use=self.__G0
                varcomp_test=self.__varcomp_test
                #varcomp_test.i_exclude=i_exclude
                
                logging.info(" (using cached null info)")
            
            #create G_exclude                         
            if (str(self.test)=="lrt_up" and result.nexclude>0):
                G_exclude = self.G_exclude(i_exclude)
            else:
                G_exclude = None            
         
            result.test = varcomp_test.testG(G1,self.test, i_exclude=i_exclude,G_exclude=G_exclude)             
            logging.info("p=%.2e",result.test['pv'])                   

            # do the permutations here, rather than as they would normally be done with a seperate call to             
            if self.nlocalperm is not None and self.nlocalperm>0:
                self.pv_from_localperm(iperm, result, t0, varcomp_test, y)
        else: #caching for 1-kernel LRT (save expensive stuff from non-permuted run)
            #currently not used, because need to be able to pass back values to work_sequence for caching
            assert iperm!=-1, "should be a permutation run"
            assert str(self.test)=="lrt", "can only use lrt, here now"
            assert self.altModel["effect"]=="mixed" and self.altModel["link"]=="linear", "caching only set up for linear lrt at the moment"
            assert self.nullModel["effect"]=="fixed", "caching only set up for one-kernel" 
            assert self.__G0 is None, "caching only 1-kernel implemented, otherwise need to permute G0 as well here"

            #permute X, y instead of G1 so can use caching                  
            yperm=y[permutationIndex]                
            Xperm=self.__X[permutationIndex]       
            result = varcomp_test.testGupdate(yperm,Xperm,self.test)  
                                                  
        t1=time.time()
        logging.info("%.2f seconds elapsed in run_test" % (t1-t0))
                
        return [result]

    def TESTBEFOREUSING_run_interactiontest(self,altset,iset,altset2,iset2,iperm = -1):
        '''
        This function does the main work of the class (when interaction tests are running).
        It is called (via a lambda) inside the loops found in 'generate_sequence'

        Input:
            altset - a set of snps
            iset - index to altset
            iperm - index to permutation (-1 means no permutation)
        Output:
            an instance of the Result class
        '''

        t0=time.time()

        self.run_once() #load files, etc. -- stuff we only want to do once no matter how many times we call 'run_test' and then 'reduce.

        logging.info("Working on permutation index #{0} out of {1} total permutations...".format(iperm, self.nperm))
        logging.info("\taltset {0}, {1} of {2}  ".format(altset, iset, len(self.altsetlist_filtbysnps)))
        self.stdout.flush()

        result = PairResult(iperm=iperm,iset=iset,setname=str(altset),iset2=iset2,setname2=str(altset2))

        #read the alternative model SNPs (for all sets) from disk.
        SNPs1a = altset.read()      #first set
        import fastlmm.util.preprocess as util
        SNPs1a['snps'] = util.standardize(SNPs1a['snps'])

        SNPs1b = altset2.read()     #second set
        SNPs1b['snps'] = util.standardize(SNPs1b['snps'])

        SNPs1=dict(SNPs1a)
        SNPs1.update(SNPs1b) #concatenate the SNP dictionaries
  
        G1 = SNPs1['snps']/sp.sqrt(SNPs1['snps'].shape[1])
        if G1.shape[1]==0: raise Exception("no SNPs to test")
        result.setsize = SNPs1['snps'].shape[1]

        logging.info(" (" + str(result.setsize) + " SNPs)")

        if self.permute >= 0 :
            permutationIndex = utilx.generatePermutation(SNPs1['snps'].shape[0],self.rseed ^ self.permute)
            G1=G1[permutationIndex]

        if iperm >= 0 :
            newseed = self.rseed ^ iperm
            if self.permute >= 0 :
                #Add a left shift so that iperm=1 and self.permute=2 gives a different newseed than iperm=2 and self.permute=1
                newseed = (newseed << 1) ^ self.permute
            permutationIndex = utilx.generatePermutation(SNPs1['snps'].shape[0], newseed)
            #permute the data to create Null-only P-values
            G1=G1[permutationIndex]

        if self.filenull is not None:
            i_exclude =  utilx.excludeinds(self.__SNPs0['data']['pos'], SNPs1['pos'], mindist = self.mindist,idist = self.idist)
            result.nexclude = i_exclude.sum()
            logging.info("numExcluded=" + str(result.nexclude))
        else:
            result.nexclude=0;

        null_model_changed=(result.nexclude>0)

        #need to make this caching smarter for when background kernel changes in every test (e.g. interactions)
        if null_model_changed:
            logging.info(" (computing needed null info anew) ")
            G0_to_use = self.__SNPs0['data']['snps'][:,~i_exclude]/SP.sqrt(self.__SNPs0['data']['snps'][:,~i_exclude].shape[1])  #excluded SNPs
            null_model=None
        else: #use cached values
            G0_to_use=self.__G0
            null_model=self.__varcomp_test           
            logging.info(" (using cached null info)")

        [result.pv,result.lik0,result.lik1] = self.test.pv_etc(filenull, G0_to_use, G1, y, x, appendBias, null_model, self.__varcomp_test, forcefullrank)

        # where is result.test used?
        result.test=self.test #probably not the best way to do this, but hacking it for now

        t1=time.time()
        logging.info("%.2f seconds elapsed" % (t1-t0))

        return result
  
    def setSNPs0(self):        
        logging.info("Reading SNPs0")
        root, ext = os.path.splitext(self.filenull)
        if (self.autoselect):
            #raise Exception("this autoselect option has been removed. please use the stand-alone FastLmm-SELECT in this same download")
            assert not self.forcefullrank, "cannot currently force full rank with autoselect=True"
            if self.extractSim is not None : raise Exception("extractSim not supported with autoselect")
            #call C++ version of FaST-LMM select to determine background Covariance matrix of the LMM
            osname = sys.platform
            dir = os.path.split(__file__)[0] #__file__ is the pathname of the file from which the module
            if (osname.find("win") >= 0):    #         was loaded, if it was loaded from a file
                fastlmmpath = os.path.join(dir,"Fastlmm_autoselect", "fastlmmc.exe")
            elif (osname.find("linux") >= 0):
                fastlmmpath = os.path.join(dir,"Fastlmm_autoselect", "fastlmmc")
                import stat
                st = os.stat(fastlmmpath)
                os.chmod(fastlmmpath, st.st_mode | stat.S_IEXEC)
            else:
                logging.info('\n\n unsupported operating system!')
            if not os.path.isfile(fastlmmpath) : raise Exception("Expect file {0}".format(fastlmmpath))
            logging.info("fastlmmC path=" + fastlmmpath)

            if (not os.path.exists('select')):
                subprocess.call("mkdir select", shell=True)

            with TemporaryFile() as stdin_temp: #An empty to file from which to read stdin (This allows this to run even when there is no normal stdin defined)
                with TemporaryFile() as stdout_temp:
                    if (self.covarfile == None):
                        subprocess.call(fastlmmpath + ' -autoselect select/auto -bfilesim ' + root + ' -pheno ' + self.phenofile + ' -pedOutFile ' + "select/null", shell=True, stdout=stdout_temp, stderr=subprocess.STDOUT, stdin=stdin_temp)
                    else:
                        subprocess.call(fastlmmpath + ' -autoselect select/auto -bfilesim ' + root + ' -pheno ' + self.phenofile + ' -covar ' + self.covarfile + ' -pedOutFile ' + "select/null", shell=True, stdout=stdout_temp, stderr=subprocess.STDOUT, stdin=stdin_temp)
                    stdout_temp.seek(0)
                    stdout_string = stdout_temp.read()
            logging.info(stdout_string)

            self.__SNPs0 = {
                "data":readPED("select/null"),
                "filename":"select/null",                
                }
            self.__SNPs0["original_iids"] = self.__SNPs0["data"]["iid"]
            self.__SNPs0["num_snps"] = self.__SNPs0["data"]["snps"].shape[1]
        else:                   
            if ext.lower() == ".bed":
                snp_set = CreateSnpSetReaderForFileNull(self.extractSim)  
                bed = Bed(root)
                snp_set_bed = snp_set.addbed(bed)
                num_snps = len(snp_set_bed)                
                self.__SNPs0={
                    "snp_set":snp_set_bed,
                    "reader":snp_set_bed.bed,
                    "filename":root,
                    #"data":readBED(root, snp_set=snp_set)
                    }
                self.__SNPs0["original_iids"] = self.__SNPs0["reader"].original_iids
                self.__SNPs0["num_snps"] = num_snps
                #import fastlmm.util.preprocess as util
                #self.__SNPs0['data']['snps'] = util.standardize(self.__SNPs0['data']['snps'])
            else:
                if self.extractSim is not None : raise Exception("extractSim not supported with filenull in ped format, please use bed format")
                self.__SNPs0 = {"data":readPED(root),
                                "filename":root}
                self.__SNPs0["original_iids"] = self.__SNPs0["data"]["iid"]
                self.__SNPs0["num_snps"] = self.__SNPs0["data"]["snps"].shape[1]

    def varcomp_test_setup(self,y,SNPs0=None, i_exclude=None): 
        #in case we exclude all SNPs (needed for score, not for lrt, not sure about lrt_up)
        nullModelTmp=self.nullModel.copy()           
        if (i_exclude is not None) and SNPs0['num_snps']==i_exclude.sum():
            excluded_all_snps = True           
            nullModelTmp['effect']='fixed'
        else:
            excluded_all_snps = False        

        if self.filenull is not None and (not excluded_all_snps): #aka two-kernel after exclusion has been done        
            varcomp_test = self.test.construct(y,self.__X, forcefullrank = self.forcefullrank,
                             SNPs0 = SNPs0, i_exclude = i_exclude, nullModel = self.nullModel, altModel = self.altModel,
                             scoring = self.scoring, greater_is_better = self.greater_is_better)
        else: #no background kernel
            varcomp_test = self.test.construct_no_backgound_kernel(y,self.__X,
                             forcefullrank = self.forcefullrank, nullModel = nullModelTmp,
                             altModel = self.altModel, scoring = self.scoring,
                             greater_is_better = self.greater_is_better)        
        return varcomp_test

    #so that it only ever gets changed once, in run_once
    @property
    def rseed(self):
        assert self._seed is not None
        return self._seed

    @property
    def mainseed(self):
        '''used to xor indexes with to get seeds for other purposes'''
        return 34343
       
    def run_once(self):
        '''
        Loading datasets (but not SNPs for alternative), etc. that we want to do only once no mater how many times we call 'run_test' and then perhaps 'reduce'.        '''         
               
        if self.calseed is not None:            
            self._seed=utilx.combineseeds(self.calseed,self.mainseed)                        
            #self._seed = self.calseed ^ self.mainseed
        else:
            self._seed = self.mainseed

        if (self._ran_once):
            return
        #self.__snapshot = self.__repr__() # must be before seeting _ran_once
        self._ran_once = True

        logging.info("Running Once")

        self._check_params_before() #must be called before test object Lrt(), Cv(), Sc() is created       
        #!! make this nicer
        if self.test=="lrt":
            self.test = Lrt()
        elif self.test=="lrt_up":
            self.test = lrt.LRT_up()
        elif self.test=="cv":
            self.test = Cv()
        elif self.test=="sc_davies":
            self.test = Sc("davies")
        elif self.test=="sc_mom":
            self.test = Sc("mom")
        else:
            raise Exception("Don't know how to construct {0}".format(self.test))        
        self._check_params_after() #must be called after test object Lrt(), Cv(), Sc() is created         

        phenhead,phentail=os.path.split(self.phenofile)
        logging.info("Reading Pheno: " +  phentail + "")        
        pheno = loadPhen(filename = self.phenofile,missing='NaN')
        pheno['vals']=pheno['vals'][:,self.mpheno-1]#use -1 so compatible with C++ version
        #pheno['header']=pheno['header'][self.mpheno-1] #header is empty 
        goodind=sp.logical_not(sp.isnan(pheno['vals']))
        pheno['vals']=pheno['vals'][goodind]
        pheno['iid']=pheno['iid'][goodind,:]        
           
        if self.genphen is not None:
            #for synthetic data generation--this is the background signal added to the phenotype
            if self.genphen["varBackNullFileGen"] is not None:
                self.__varBackNullSnpsGen=readBED(self.genphen["varBackNullFileGen"])
                import fastlmm.util.preprocess as util
                self.__varBackNullSnpsGen['snps'] = util.standardize(self.__varBackNullSnpsGen['snps'])
            else:
                self.__varBackNullSnpsGen=None

        if self.filenull is not None:
            self.setSNPs0()
            logging.info("Filter SNPs0")
            #prelim intersection so don't read in all null SNP iids if way more than in pheno; now redundant
            #filter(pheno, self.__SNPs0['data'])
        else:
            if self.extractSim is not None : raise Exception("extractSim should be specified only if filenull is specified.")
            self.__SNPs0=None
            #self.__G0=None

        if self.covarfile==None:
            covar=None
        else:
            covar=loadPhen(self.covarfile,missing='NaN')
            if self.covarimp=='standardize':
                covar['vals'],fracmissing=utilx.standardize_col(covar['vals'])
            elif self.covarimp is None:
                pass
            else:
                raise Exception("covarimp=" + self.covarimp + " not implemented")          

        covar,pheno,indarr=self.intersect_data(covar, pheno)
        N=pheno['vals'].shape[0];
        if self.covarfile==None:
            self.__X = sp.ones((N,1))
        else:
            #check for covariates which are constant, as the lmm code crashes on these            
            badind = utilx.indof_constfeatures(covar['vals'], axis=0)
            if len(badind)>0:
                raise Exception("found constant covariates with indexes: %s: please remove, or modify code here to do so" % badind)
            self.__X=sp.hstack((sp.ones((N,1)),covar['vals']))
        self.__y = pheno['vals']

        if not covar is None and sp.isnan(covar['vals'].sum()): raise Exception("found missing values in covariates file that remain after intersection")
        if sp.isnan(self.__y.sum()): raise Exception("found missing values in phenotype file that remain after intersection")
        #if hasattr(self,'__G0') and not self.__G0 is None and sp.isnan(self.__G0.sum())): raise Exception("found missing values in background SNPs that remain after intersection")

        #creating sets from set defn files. Looks at bed file to filter down the SNPs to only those present in the bed file
        self.altset_list,self.altsetlist_filtbysnps = self.create_altsetlist_filtbysnps(self.altset_list,self.alt_snpreader)
        if self.altset_list2 is not None:
                self.altset_list2,self.__altsetlist2_filtbysnps = self.create_altsetlist_filtbysnps(self.altset_list2,self.alt_snpreader)

        if self.npermabs is not None and self.nperm is not None:
            assert self.nperm is None, "cannot use both nperm and npermabs"
        if self.npermabs is not None:
            num_tests = len(self.altsetlist_filtbysnps)
            self.nperm = int(sp.ceil(self.npermabs/num_tests))
            logging.info("using npermabs=%i, found %i tests, so using %i permutations per test" % (self.npermabs, num_tests, self.nperm))            
            self.npermabs = None
        
        logging.info("------------------------------------------------")
        logging.info("Found " + str(self.__X.shape[0]) +  " individuals")
        logging.info("Found " + str(self.__X.shape[1]) +  " covariates (including bias)")
        if self.__SNPs0 is not None:
            logging.info("Found " + str(self.__SNPs0['num_snps']) +  " SNPs for null kernel")
        if (len(self.__y.shape)>1): nPhen=self.__y.shape[1]
        else: nPhen=1
        logging.info("Found " + str(nPhen) +  " phenotypes")
        if self.filenull != None:
            logging.info("Running test " +  str(self.test) + " with 1 background kernel")
        else:
            logging.info("Running test " +  str(self.test)  + " with no background kernel")
        #CL3_28_14:The following never evaluates True, as it should be #if str(self.test)=="lrt":
        #but the body is buggy, so it would crash
        #if self.test=="lrt":
        #    logging.info("nullmodel: " + "link=" + self.nullModel['link'] + ", " 
        #                          + "approx=" + self.nullModel['approx'] + ", "
        #                          + "pen="+ self.nullModel['penalty'])
        #    logging.info("altmodel: " + "link=" + self.altModel['link'] + ", " 
        #                          + "approx=" + self.altModel['approx'] + ", "
        #                          + "pen="+ self.altModel['penalty'])        
        logging.info("Found " + str(self.work_count/(self.nperm+1)) + " tests to run (" + str(self.work_count) + " work items, including " + str(self.nperm) + " permutations)")
        if self.nullfitfile is not None:
            logging.info("nullfitfile=" + self.nullfitfile)
        if self.altset_list2 is not None:
            logging.info("(running paired set tests)")
        else:
            logging.info("(running single set tests)")        
        if self.minsetsize is not None and self.maxsetsize is not None:
            logging.info("min set size=%i, max set size=%i", self.minsetsize, self.maxsetsize) 
        if self.mpheno is not None: logging.info("mpheno=%i", self.mpheno)
        if self.extractSim is not None: logging.info("extractSim=%s",self.extractSim)
        logging.info("------------------------------------------------")

        logging.info("Creating null model... ")
        #cache whatever is needed for each case (sometimes likelihood, sometimes rotated items, sometimes nothing)
        t0=time.time()
        
        #if self.filenull is not None: #aka two-kernel
        #    self.__G0 = self.__SNPs0['data']['snps']/sp.sqrt(self.__SNPs0['data']['snps'].shape[1])        
        #else:
        #    self.__G0=None
        
        #for some models, this caches important and expensive information
        #as appropriate, gets re-computed in run_test()        
        if self.genphen is None: #or self.genphen['once']: THIS CAUSES REAL DATA TO GET USED         
            #self.__varcomp_test=self.varcomp_test_setup(self.__y,self.__G0)            
            self.__varcomp_test=self.varcomp_test_setup(self.__y,self.__SNPs0)
        #elif self.genphen is not None and self.__G0 is not None:
        elif self.genphen is not None and self.__SNPs0 is not None:
            raise Exception("this will run if you comment out this line, but I think will be inefficient because it's not caching null info across each gene-set specific phenotypes even when G0 remains constant across these")
            self.__varcomp_test=None
        else:
            # single kernel synthetic, fine to re-do null model computations each time
            self.__varcomp_test=None

        t1=time.time()
        logging.info("done. %.2f seconds elapsed" % (t1-t0))

    def check_id_order(self, covar, pheno):        
        ids_pheno=pheno['iid']
        ids_SNPs=self.alt_snpreader.original_iids
        if (len(ids_pheno)!=len(ids_SNPs)) or (not sp.all(ids_pheno==ids_SNPs)):
          return False
        if self.covarfile is not None:
            ids_X=covar['iid']
            if (len(ids_pheno)!=len(ids_X)) or (not sp.all(ids_pheno==ids_X)): return False
        if self.filenull is not None:
            ids_nullSNPs=self.__SNPs0['original_iids']
            if not sp.all(ids_pheno==ids_nullSNPs): return False
        if hasattr(self,"__varBackNullSnpsGen") and self.__varBackNullSnpsGen is not None: #for synthetic data generation            
            ids_SNPgen=self.__varBackNullSnpsGen["iid"]
            if not sp.all(ids_pheno==ids_SNPgen): return False
        return True

    def intersect_data(self, covar, pheno):
        if self.check_id_order(covar,pheno):
            indarr = SP.arange(pheno['iid'].shape[0])
            logging.info(str(indarr.shape[0]) + " IDs match up across data sets")
        else:            
            logging.info("IDs do not match up, so intersecting the data over individuals")            
            if self.filenull is not None:
                nullsnpids=self.__SNPs0['original_iids']
            else:
                nullsnpids=None
            if self.covarfile is not None:
                covarids=covar['iid']
            else:
                covarids=None
            if hasattr(self,"__varBackNullSnpsGen") and self.__varBackNullSnpsGen is not None:
                varbacknullsnp=self.__varBackNullSnpsGen['iid']
            else:
                varbacknullsnp=None

            #the order of inputs here is reflected in the reordering indexes below of 0,1,2,3,4
            import warnings
            #!!warnings.warn("This intersect_ids is deprecated. Pysnptools includes newer versions of intersect_ids", DeprecationWarning)
            indarr=utilx.intersect_ids([pheno['iid'],self.alt_snpreader.original_iids,covarids,nullsnpids,varbacknullsnp])            
            assert indarr.shape[0]>0, "no individuals remain after intersection, check that ids match in files"
            #sort the indexes so that SNPs ids in their original order (and 
            #therefore we have to move things around in memory the least amount)
            #[indarr[:,3] holds the SNP order    
            sortind=sp.argsort(indarr[:,3])
            indarr=indarr[sortind]

            pheno['iid']=pheno['iid'][indarr[:,0]]
            pheno['vals']=pheno['vals'][indarr[:,0]]

            self.alt_snpreader.ind_used=indarr[:,1]

            if self.covarfile is not None:
                covar['iid']=covar['iid'][indarr[:,2]]
                covar['vals']=covar['vals'][indarr[:,2]]

            if self.filenull is not None:  
                #copy in memory to make later access faster (presumably?):
                #self.__SNPs0['iid']=sp.copy(self.__SNPs0['iid'][indarr[:,3]])
                #self.__SNPs0['snps']=sp.copy(self.__SNPs0['snps'][indarr[:,3]])
                #self.__SNPs0['data']['iid']=self.__SNPs0['data']['iid'][indarr[:,3]]
                if self.__SNPs0.has_key('data'):
                    self.__SNPs0['data']['iid'] = self.__SNPs0['data']['iid'][indarr[:,3]]
                    self.__SNPs0['data']['snps'] = self.__SNPs0['data']['snps'][indarr[:,3]]
                else:
                    self.__SNPs0['reader'].ind_used = indarr[:,3]
                #self.__SNPs0['data']['snps']=self.__SNPs0['data']['snps'][indarr[:,3]]

            if hasattr(self,"_varBackNullSnpsGen") and self.__varBackNullSnpsGen is not None:
                self.__varBackNullSnpsGen['iid']=self.__varBackNullSnpsGen['iid'][indarr[:,4]]
                self.__varBackNullSnpsGen['snps']=self.__varBackNullSnpsGen['snps'][indarr[:,4]]

            logging.info(str(indarr.shape[0]) + " ids left")
        
        return covar,pheno,indarr


    def create_altsetlist_filtbysnps(self, altset_list, altsetbed):
        '''
        returns: altset_list, altsetlist_filtbysnps
        'altset_list' is the raw set defintion read in from file
        'altsetlist_filtbysnps' is the result of filtering altset_list by SNPsin the specified altsetbed SNP file
        Additionally, only allows sets specified by sets are allowed (if this is not None)
        '''
        if isinstance(altset_list, str):
            altset_list = SnpAndSetNameCollection(altset_list)

        if self.sets is not None:  #filter by those sets specified on the command line, stored in self.sets
            altset_list=Subset(altset_list, self.sets)

        if self.minsetsize is not None or self.maxsetsize is not None:
            altset_list=MinMaxSetSize(altset_list,minsetsize=self.minsetsize,maxsetsize=self.maxsetsize) #would be nice if applying 1,None wouldn't wrap at all (becuase that is the identity)

        altsetlist_filtbysnps = altset_list.addbed(altsetbed)

        if len(altsetlist_filtbysnps)==0: raise Exception("Expect altset_list to contain at least one set")

        return altset_list, altsetlist_filtbysnps

    # overridden to provide a nice name on the cluster.
    def __str__(self):
        if self.outfile == None:
            return self.__class__.__name__
        else:
            return "{0} {1}".format(self.__class__.__name__, self.outfile)

    def __repr__(self):
        import cStringIO
        fp = cStringIO.StringIO()
        fp.write("{0}(\n".format(self.__class__.__name__))
        varlist = []
        for f in dir(self):
            if f.startswith("_"): # remove items that start with '_'
                continue
            if type(self.__class__.__dict__.get(f,None)) is property: # remove @properties
                continue
            if callable(getattr(self, f)): # remove methods
                continue
            varlist.append(f)
        for var in varlist[:-1]: #all but last
            fp.write("\t{0}={1},\n".format(var, getattr(self, var).__repr__()))
        var = varlist[-1] # last
        fp.write("\t{0}={1})\n".format(var, getattr(self, var).__repr__()))
        result = fp.getvalue()
        fp.close()
        return result

def CreateSnpSetReaderForFileNull(snp_set):
    if (snp_set is None):
        return AllSnps()
    if isinstance(snp_set, str):        
        return SnpsFromFile(snp_set)
    return snp_set



if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    logging.info("last modified: %s" % time.ctime(os.path.getmtime(__file__)))
    logging.info("-------------------------------------------------")       
    
    # Parse the command line
    if len(sys.argv) == 2:
        sys.argv.append("Local()")
    if len(sys.argv) != 3:
        logging.info("Usage:python FastLmmSet.py <inputfile> [optional <runner>]")
        sys.exit(1)
    file,infile,runner_string = sys.argv
    if not os.path.exists(infile): raise Exception("Fatal Error: " + infile + " does not exist")

    #Run the job specified in infile    
    from fastlmm.association.FastLmmSet import *
    exec("runner = " + runner_string)
    d = open(infile).read()    
    exec(d)     
          
    tt0=time.time()    
    runner.run(distributable)    
    tt1=time.time()
    logging.info("Final elapsed time for all processing is %.2f seconds" % (tt1-tt0))




import scipy as SP
import scipy.stats as ST
import numpy as NP
from numpy import dot
from scipy.linalg import cholesky,solve_triangular
from fastlmm.external.util.math import check_definite_positiveness,check_symmetry,ddot,dotd,trace2
from fastlmm.external.util.math import stl, stu
from sklearn.base import BaseEstimator
from fastlmm import Pr
import sys

'''
    Important! Always run test.py in the current folder for unit testing after
    changes have been made.
'''

class GLMM(object):
    '''
    generalized linear mixed model having up to two linear kernels
    f ~ N(X*beta, sig02*K0 + sig12*K1 + sign2*I),
    y ~ Bern(link(f))
    where
    K0 = G0*G0^T
    K1 = G1*G1^T
    '''

    def __init__(self, penalty=None, penalizeBias=False, debug=False):
        '''
        Input:
        approx        : 'laplace' or 'ep'
        link          : 'logistic' or 'erf'
        penalty       : None, 'l1' or 'l2'
        penalizeBias  : True or False
        '''
        self._debug = debug
        self._X=None
        self._y=None
        self._sig02=0.0
        self._sig12=0.0
        self._sign2=0.0
        self._G0 = None
        self._G1 = None
        self._isK0Set=False
        self._isK1Set=False
        self._beta=None
        self._mapIn2OutY=None
        self._mapOut2InY=None
        self._outYType=None
        assert penalty in set([None,'l1','l2'])
        assert penalty != 'l1', 'l1 penalizer is not fully implemented yet.'
        self._penalty = penalty
        self._penalizeBias = penalizeBias
        self._hasBias = False
        self._biasIndex = None
        # penalization weights
        self._lambdaBeta = 0.1
        self._lambdaSig02 = 0.1
        self._lambdaSig12 = 0.1
        self._lambdaSign2 = 0.1

        self._updateConstantsCount = 1
        self._updateApproximationCount = 1

    def setX(self, X):
        '''
        set the fixed effects X (covariates).
        --------------------------------------------------------------------------
        Input:
        X       : [N*D] 2-dimensional array of covariates
        --------------------------------------------------------------------------
        '''
        assert type(X) is NP.ndarray, 'X must be a numpy.ndarray.'
        assert NP.all(NP.isfinite(X) & ~NP.isnan(X)), 'X must have only numbers.'
        assert len(X.shape)==2, 'X must be a 2-dimensional array.'
        assert X.shape[0] > 0
        assert X.shape[1] > 0
        self._X = X
        self._updateConstantsCount += 1

        self._hasBias = False
        for i in xrange(X.shape[1]):
            if len(NP.unique(X[:,i])) == 0 and X[0,i] == 1.0:
                self._hasBias = True
                self._biasIndex = i
                break

    def sety(self, y):
        '''
        set the phenotype y.
        --------------------------------------------------------------------------
        Input:
        y       : [N] 1-dimensional array of phenotype values (-1.0 or 1.0)
        --------------------------------------------------------------------------
        '''
        assert type(y) is NP.ndarray, 'y must be a numpy.ndarray.'
        assert len(y.shape)==1, 'y must be a 1-dimensional array.'
        assert y.shape[0] > 0

        uniquey = list(set(y))
        assert len(uniquey)==2, 'y must have two unique values.'
        uniquey = sorted(uniquey)

        self._mapIn2OutY = {-1.0:uniquey[0],+1.0:uniquey[1]}
        self._mapOut2InY = {uniquey[0]:-1.0,uniquey[1]:+1.0}
        self._outYType = y.dtype

        self._y = NP.empty(y.shape[0])
        self._y[y==uniquey[0]] = -1.0
        self._y[y==uniquey[1]] = +1.0

        self._updateConstantsCount += 1

    def setG(self, G0, G1=None):
        '''
        set the kernels K0 and K1 from G0 and G1.
        ----------------------------------------------------------------------------
        Input:
        G0              : [N*k0] array of random effects
        G1              : [N*k1] array of random effects (optional)
        -----------------------------------------------------------------------------
        '''
        assert type(G0) is NP.ndarray, 'G0 must be a numpy.ndarray.'
        assert NP.all(NP.isfinite(G0) & ~NP.isnan(G0)), 'G0 must have only numbers.'
        assert len(G0.shape)==2, 'G0 must be a 2-dimensional array.'
        assert G0.shape[0] > 0
        assert G0.shape[1] > 0
        self._G0 = G0

        if G1 is not None:
            assert type(G1) is NP.ndarray, 'G1 must be a numpy.ndarray.'
            assert NP.all(NP.isfinite(G1) & ~NP.isnan(G1)), 'G1 must have only numbers.'
            assert len(G1.shape)==2, 'G1 must be a 2-dimensional array.'
            assert G1.shape[0] > 0
            assert G1.shape[1] > 0
            self._G1 = G1

        self._isK0Set = G0 is not None
        self._isK1Set = G1 is not None

        self._updateConstantsCount += 1

    # Operation dot(A,K).
    def _ldotK(self, A):
        pass

    # Operation dot(K,A).
    def _rdotK(self, A):
        pass

    # Return K's diagonal.
    def _dKn(self):
        pass

    def printDebug(self):
        pass

    '''
    -------------------------------------------------------------------------------
    Begin of hyperparameters
    '''
    @property
    def sig02(self):
        return self._sig02

    @sig02.setter
    def sig02(self, v):
        assert NP.isscalar(v)
        assert NP.isfinite(v)
        if self._sig02 == v:
            return
        self._updateApproximationCount += 1
        self._sig02 = v

    @property
    def sig12(self):
        return self._sig12

    @sig12.setter
    def sig12(self, v):
        assert NP.isscalar(v)
        assert NP.isfinite(v)
        if self._sig12 == v:
            return
        self._updateApproximationCount += 1
        self._sig12 = v

    @property
    def sign2(self):
        return self._sign2

    @sign2.setter
    def sign2(self, v):
        assert NP.isscalar(v)
        assert NP.isfinite(v)
        if self._sign2 == v:
            return
        self._updateApproximationCount += 1
        self._sign2 = v

    @property
    def beta(self):
        return self._beta.copy()

    @beta.setter
    def beta(self, v):
        assert NP.all(NP.isfinite(v))
        if self._beta is not None and NP.all(self._beta == v):
            return
        self._updateApproximationCount += 1
        self._beta = v.copy()
    '''
    End of hyperparameters
    -------------------------------------------------------------------------------
    '''

    def _xMeanCov(self, xX, xG0, xG1):
        '''
        Computes the mean and covariance between the latent variables.
        You can provide one or n latent variables.
        ----------------------------------------------------------------------------
        Input:
        xX              : [D] array of fixed effects or
                          [n*D] array of fixed effects for each latent variable
        xG0             : [k0] array of random effects or
                          [n*k0] array of random effects for each latent variable
        xG1             : [k1] array of random effects or
                          [n*k1] array of random effects for each latent variable
        -----------------------------------------------------------------------------
        Output:
        xmean           : [n*D] means of the provided latent variables
        xK01            : [N] or [n*N] covariance between provided and prior latent
                          variables
        xkk             : float or [n] covariance between provided latent variables 
        -----------------------------------------------------------------------------
        '''
        xmean = xX.dot(self._beta)

        if xG0 is not None:
            xK01 = self._sig02*(xG0.dot(self._G0.T))
        else:
            if len(xX.shape)==1:
                xK01 = NP.zeros(self._N)
            else:
                xK01 = NP.zeros((xX.shape[0], self._N))

        if len(xG0.shape)==1:
            xkk = self._sig02 * xG0.dot(xG0) + self._sign2
        else:
            xkk = self._sig02 * dotd(xG0,xG0.T) + self._sign2

        if xG1 is not None:
            xK01 += self._sig12*(xG1.dot(self._G1.T))
            if len(xG0.shape)==1:
                xkk += self._sig12 * xG1.dot(xG1)
            else:
                xkk += self._sig12 * dotd(xG1,xG1.T)

        return (xmean,xK01,xkk)

    def _predict(self, xmean, xK01, xkk, prob):
        pass

    def predict(self, xX, xG0, xG1=None, prob=True):
        '''
        Compute the probability of y=1.0, for each provided latent variable.
        It can instead return only the most probable class by setting prob=False.
        --------------------------------------------------------------------------
        Input:
        xX              : [D] array of fixed effects or
                          [n*D] array of fixed effects for each latent variable
        xG0             : [k0] array of random effects or
                          [n*k0] array of random effects for each latent variable
        xG1             : [k1] array of random effects or
                          [n*k1] array of random effects for each latent variable
        prob            : whether you want the probabilities or not.
        --------------------------------------------------------------------------
        Output:
        Array of probabilities or the most probable classes.
        --------------------------------------------------------------------------
        '''
        self._updateConstants()

        if len(xX.shape)==1:
            assert xX.shape[0]==self._D
            assert xG0.shape[0]==self._G0.shape[1]
            if xG1 is not None:
                assert xG1.shape[0]==self._G1.shape[1]

        elif len(xX.shape)==2:
            assert xX.shape[1]==self._D
            assert xG0.shape[1]==self._G0.shape[1]
            if xG1 is not None:
                assert xG1.shape[1]==self._G1.shape[1]

        else:
            assert False

        (xmean,xK01,xkk) = self._xMeanCov(xX, xG0, xG1)

        ps = self._predict(xmean, xK01, xkk, prob)
        if prob==False:
            return self._in2outy(ps)
        return ps

    def _in2outy(self, yin):
        '''
        Translates the labels used in this class (-1.0, 1.0) to the labels
        being used by the user.
        '''
        yout = NP.empty(yin.shape[0], dtype=self._outYType)
        yout[yin == -1] = self._mapIn2OutY[-1]
        yout[yin == +1] = self._mapIn2OutY[+1]
        return yout

    def _updateApproximation(self):
        pass

    def optimize(self, optSig02=True, optSig12=True, optSign2=True, optBeta=True):
        '''
        Minimize the cost, which is the negative of marginal loglikelihood plus the
        penalty, by adjusting the hyperparameters.
        --------------------------------------------------------------------------
        Input:
        optSig02, optSig12, optSign2, and optSigBeta can be True or False and
        are used to choose which hyperparameters are going to be optimized.
        --------------------------------------------------------------------------
        '''
        self._updateConstants()

        tol = 1e-6
        niterNoImprov = 5
        lowerBound = 1e-6

        betaScale = 4.0

        # initial values
        if optSig02:
            self.sig02 = 1.0
        if optSig12:
            self.sig12 = lowerBound
        if optSign2:
            self.sign2 = lowerBound
        if optBeta:
            self.beta = NP.zeros(self._D)
        
        bestCost = self._optimize(lowerBound, optSig02, optSig12, optSign2, optBeta)
        bestSolution = self._wrap_hyp(True, True, True, True)

        i = 0
        ntries = 0
        while i < niterNoImprov:
            if optSig02:
                self.sig02 = max(NP.random.chisquare(1), lowerBound)
            if optSig12:
                self.sig12 = max(NP.random.chisquare(1), lowerBound)
            if optSign2:
                self.sign2 = max(NP.random.chisquare(1), lowerBound)
            if optBeta:
                self.beta = NP.random.normal(scale=betaScale, size=self._D)

            cost = self._optimize(lowerBound, optSig02, optSig12, optSign2, optBeta)
            if cost < bestCost:
                bestCost = cost
                bestSolution = self._wrap_hyp(True, True, True, True)
                if abs(cost-bestCost) > tol:
                    i = 0
                else:
                    i += 1
            else:
                i += 1
            ntries += 1

        #if self._verbose:
        #    Pr.prin('Number of tries: {}.'.format(ntries))

        self._unwrap_hyp(bestSolution, True, True, True, True)

        # If K0 is null, marginal likelihood is independent on
        # self._sig02. Thus, let it be 0.0 by default.
        if not self._isK0Set:
            self.sig02 = 0.0

        # If K1 is null, marginal likelihood is independent on
        # self._sig12. Thus, let it be 0.0 by default.
        if not self._isK1Set:
            self.sig12 = 0.0

        self._check_sigmas_at_zero(lowerBound, optSig02, optSig12, optSign2, optBeta)
        
    def _optimize(self, lowerBound, optSig02=True, optSig12=True, optSign2=True, optBeta=True):
        def func(x):
            self._unwrap_hyp(x, optSig02, optSig12, optSign2, optBeta)
            return -self.marginal_loglikelihood()

        def grad(x):
            self._unwrap_hyp(x, optSig02, optSig12, optSign2, optBeta)
            return -self._mll_gradient()

        x = self._wrap_hyp(optSig02, optSig12, optSign2, optBeta)

        # bounds on sig02, sig12, sign2, and w
        bounds = []
        maxsig2 = 50.0
        maxbeta = 50.0
        if optSig02:
            bounds.append((lowerBound,maxsig2))
        if optSig12:
            bounds.append((lowerBound,maxsig2))
        if optSign2:
            bounds.append((lowerBound,maxsig2))
        if optBeta:
            bounds += [(-maxbeta,maxbeta)]*self._D

        (xfinal,aa,bb) = SP.optimize.fmin_tnc(func, x, fprime=grad, bounds=bounds, disp=0)
#        (xfinal,cost,msgs) = SP.optimize.fmin_l_bfgs_b(func, x, fprime=grad, bounds=bounds,
#                                                        disp=self._verbose)
#                                                        disp=False)
        self._unwrap_hyp(xfinal, optSig02, optSig12, optSign2, optBeta)
        marg = self.marginal_loglikelihood()
#        marg = -cost
        
        #if self._verbose:
        #    Pr.prin('Best hyperparameters. sig02 '+str(self._sig02)+' sig12 '+str(self._sig12)+' sign2 '+str(self._sign2)+' beta '+str(self._beta))
        #    Pr.prin('marg '+str(marg))

        return -marg

    def _check_sigmas_at_zero(self, lowerBound, optSig02, optSig12, optSign2, optBeta):
        pmargll = self.marginal_loglikelihood()

        psig02 = self._sig02
        if optSig02:
            if abs(self._sig02-lowerBound)<1e-7:
                self.sig02 = 0.0

        psig12 = self._sig12
        if optSig12:
            if abs(self._sig12-lowerBound)<1e-7:
                self.sig12 = 0.0

        psign2 = self._sign2
        if optSign2:
            if abs(self._sign2-lowerBound)<1e-7:
                self.sign2 = 0.0

        margll = self.marginal_loglikelihood()

        if margll < pmargll:
            self.sig02 = psig02
            self.sig12 = psig12
            self.sign2 = psign2

    def _is_kernel_zero(self):
        return self._sig02==0.0 and self._sig12==0.0 and self._sign2==0.0
    
    def marginal_loglikelihood(self):
        '''
        Returns (regular marginal loglikelihood) - penalty
        '''
        margll = self._regular_marginal_loglikelihood()

        if self._penalty is None:
            return margll

        if self._penalizeBias:
            assert self._hasBias, "You set to penalize bias but there isn't one."
            beta = self.beta
        else:
            beta = self._betaNoBias()

        if self._penalty == 'l1':
            assert False, 'Not implemented yet.'
            return margll - self._lambdaBeta * sum(abs(beta)) - self._lambdaSign2 * abs(self.sign2)

        elif self._penalty == 'l2':
            #return margll - self._lambdaBeta * NP.dot(beta,beta) - self._lambdaSign2 * self.sign2**2
            r = margll - self._lambdaBeta * NP.dot(beta,beta) - self._lambdaSign2 * self.sign2**2
            if self._isK0Set:
                r -= self._lambdaSig02 * self.sig02**2
            if self._isK1Set:
                r -= self._lambdaSig12 * self.sig12**2
            return r

        assert False, 'Unknown penalty.'

    def _betaNoBias(self):
        beta = self.beta.copy()
        if self._hasBias:
            assert self._hasBias, "You set to penalize bias but there isn't one."
            beta = NP.concatenate( (beta[:self._biasIndex], beta[self._biasIndex+1:]) )
        return beta

    def _mll_gradient(self, optSig02=True, optSig12=True, optSign2=True, optBeta=True):
        '''
        Marginal loglikelihood gradient.
        '''
        g = self._rmll_gradient()
        if self._penalty is None or optBeta is False:
            return g

        if optSig02:
            sig02Index = 0
        if optSig12:
            sig12Index = 0+optSig02
        if optSign2:
            sign2Index = optSig02 + optSig12

        firstIndex = optSig02 + optSig12 + optSign2
        
        gs = NP.zeros(g.shape[0])
        if self._penalty == 'l1':
            assert False, 'Not implemented yet.'

        elif self._penalty == 'l2':

            gs[firstIndex:] = 2.0 * self._lambdaBeta * self.beta

            if optSign2:
                gs[sign2Index] = 2.0 * self._lambdaSign2 * self.sign2

            if self._isK0Set and optSig02:
                gs[sig02Index] = 2.0 * self._lambdaSig02 * self.sig02
            if self._isK1Set and optSig12:
                gs[sig12Index] = 2.0 * self._lambdaSig12 * self.sig12
        else:
            assert False, 'Unknown penalty.'

        return g - gs

    def _wrap_hyp(self, optSig02, optSig12, optSign2, optBeta):
        x = NP.empty(sum([optSig02, optSig12, optSign2]) + sum([optBeta])*self._D)
        i = 0
        if optSig02:
            x[i] = self._sig02
            i += 1
        if optSig12:
            x[i] = self._sig12
            i += 1
        if optSign2:
            x[i] = self._sign2
            i += 1
        if optBeta:
            x[i:] = self._beta
        return x

    def _unwrap_hyp(self, x, optSig02, optSig12, optSign2, optBeta):
        i = 0
        if optSig02:
            self.sig02 = x[i]
            i += 1
        if optSig12:
            self.sig12 = x[i]
            i += 1
        if optSign2:
            self.sign2 = x[i]
            i += 1
        if optBeta:
            self.beta = x[i:]

    def _calculateMean(self):
        return self._X.dot(self._beta)

    def _updateConstants(self):
        pass

class GLMM_N3K1(GLMM):
    def __init__(self, penalty=None, penalizeBias=False, debug=False):
        GLMM.__init__(self, penalty=penalty, penalizeBias=penalizeBias, debug=debug)
        self._K0 = None
        self._K1 = None

    def setK(self, K0, K1=None):
        '''
        set the Kernels K0 and K1.
        --------------------------------------------------------------------------
        Input:
        K0 : [N*N] array, random effects covariance (positive definite)
        K1 : [N*N] array, random effects covariance (positive definite)(optional)
        --------------------------------------------------------------------------
        '''
        assert type(K0) is NP.ndarray, 'K0 must be a numpy.ndarray.'
        assert NP.all(NP.isfinite(K0) & ~NP.isnan(K0)), 'K0 must have only numbers.'
        assert len(K0.shape)==2, 'K0 must be a 2-dimensional array.'
        assert K0.shape[0] > 0
        assert K0.shape[1] > 0
        assert check_symmetry(K0), 'K0 must be a symmetric matrix.'
        assert check_definite_positiveness(K0), 'K0 must be a definite positive matrix.'
        self._K0 = K0

        if K1 is not None:
            assert type(K1) is NP.ndarray, 'K1 must be a numpy.ndarray.'
            assert NP.all(NP.isfinite(K1) & ~NP.isnan(K1)), 'K1 must have only numbers.'
            assert len(K1.shape)==2, 'K1 must be a 2-dimensional array.'
            assert K1.shape[0] > 0
            assert K1.shape[1] > 0
            assert check_symmetry(K1), 'K1 must be a symmetric matrix.'
            assert check_definite_positiveness(K1), 'K1 must be a definite positive matrix.'
            self._K1 = K1

        self._isK0Set = K0 is not None
        self._isK1Set = K1 is not None

        self._updateConstantsCount += 1

    def _updateConstants(self):
        '''
        Updates some constant members, which is needed when the user, for e.g., change
        the effects.
        '''
        if self._updateConstantsCount == 0:
            return

        assert self._X is not None, 'You must set X.'

        if self._G0 is not None and self._K0 is not None:
            assert abs(self._K0-self._G0.dot(self._G0.T)).max() < NP.sqrt(NP.finfo(NP.float).eps), 'You have set both G0 and K0, but K0!=G0*G0^T.'

        if self._G1 is not None and self._K1 is not None:
            assert abs(self._K1-self._G1.dot(self._G1.T)).max() < NP.sqrt(NP.finfo(NP.float).eps), 'You have set both G1 and K1, but K1!=G1*G1^T.'

        self._N = self._X.shape[0]
        self._D = self._X.shape[1]

        if self._K0 is None:
            if self._G0 is None:
                self._K0 = NP.zeros([self._N, self._N])
            else:
                self._K0 = self._G0.dot(self._G0.T)

        if self._K1 is None:
            if self._G1 is None:
                self._K1 = NP.zeros([self._N, self._N])
            else:
                self._K1 = self._G1.dot(self._G1.T)

        assert self._K0.shape[0]==self._X.shape[0], 'K0 (or G0) and X must have the same number of rows.'
        assert self._K1.shape[0]==self._X.shape[0], 'K1 (or G1) and X must have the same number of rows.'

        assert self._y is not None, 'You must set y.'
        assert self._y.shape[0]==self._X.shape[0], 'X and y have incompatible sizes.'
        
        self._updateApproximationCount += 1
        self._updateConstantsCount = 0

    # If A is dxN, we have O(dN^2) operations.
    def _ldotK(self, A):
        #TODO: pre-calculate in updateApproximationBegin
        R = self._sig02 * dot(A, self._K0) + A * self._sign2
        if self._K1 is not None:
            R += self._sig12 * dot(A, self._K1)
        return R

    # If A is Nxd, we have O(dN^2) operations.
    def _rdotK(self, A):
        #TODO: pre-calculate in updateApproximationBegin
        R = self._sig02 * dot(self._K0, A) + A * self._sign2
        if self._K1 is not None:
            R += self._sig12 * dot(self._K1, A)
        return R

    # Return K's diagonal, in O(N).
    def _dKn(self):
        #TODO: pre-calculate in updateApproximationBegin
        d = self._sig02 * NP.diag(self._K0) + self._sign2
        if self._K1 is not None:
            d += self._sig12 * NP.diag(self._K1)
        return d

    def _calculateLn(self, K, D):
        Bn = ddot(D, ddot(K, D, left=False), left=True)
        Bn[NP.diag_indices_from(Bn)] += 1.0
        Ln = cholesky(Bn, lower=True, check_finite=False)
        return Ln


class GLMM_N1K3(GLMM):
    def __init__(self, penalty=None, penalizeBias=False, debug=False):
        GLMM.__init__(self, penalty=penalty, penalizeBias=penalizeBias, debug=debug)

    # If A is dxN, we have O(kdN) operations (assuming N>=k>=d).
    def _ldotK(self, A):
        R = self._sig02 * (A.dot(self._G0).dot(self._G0.T)) + A * self._sign2
        if self._G1 is not None:
            R += self._sig12 * (A.dot(self._G1).dot(self._G1.T))
        return R

    # If A is Nxd, we have O(kdN) operations (assuming N>=k>=d).
    def _rdotK(self, A):
        R = self._sig02 * (self._G0.dot(self._G0.T.dot(A))) + A * self._sign2
        if self._G1 is not None:
            R += self._sig12 * (self._G1.dot(self._G1.T.dot(A)))
        return R

    # Return K's diagonal, in O((k0+k1)*N).
    def _dKn(self):
        d = self._sig02 * dotd(self._G0, self._G0.T) + self._sign2
        if self._G1 is not None:
            d += self._sig12 * dotd(self._G1, self._G1.T)
        return d

    def _updateConstants(self):
        '''
        Updates some constant members, which is needed when the user, for e.g., change
        the effects.
        '''
        if self._updateConstantsCount == 0:
            return

        assert self._X is not None, 'You must set X.'

        self._N = self._X.shape[0]
        self._D = self._X.shape[1]

        if self._G0 is None:
            self._G0 = NP.zeros([self._N, 1])

        if self._G1 is None:
            self._G1 = NP.zeros([self._N, 1])

        assert self._G0.shape[0]==self._X.shape[0], 'G0 and X must have the same number of rows.'
        assert self._G1.shape[0]==self._X.shape[0], 'G1 and X must have the same number of rows.'

        assert self._y is not None, 'You must set y.'
        assert self._y.shape[0]==self._X.shape[0], 'X and y have incompatible sizes.'

        self._updateApproximationCount += 1
        self._updateConstantsCount = 0

    def _calculateG01(self):
        G01 = NP.concatenate( (NP.sqrt(self._sig02)*self._G0, NP.sqrt(self._sig12)*self._G1), axis=1 )
        return G01

    def _calculateLk(self, G01, D):
        Bk  = dot(G01.T, ddot(D, G01, left=True))
        Bk[NP.diag_indices_from(Bk)] += 1.0
        Lk = cholesky(Bk, lower=True, check_finite=False)
        return Lk



import os
import cPickle as pickle
import subprocess, sys, os.path
from fastlmm.util.runner import *
import logging

class IndirectDistributable(object): #implements IDistributable

    def __init__(self, indirect_runner, distributable_list=None):
        self.indirect_runner = indirect_runner
        if distributable_list is None:
            distributable_list = []

        self._distributable_list = distributable_list

    def append(self, distributable):
        self._distributable_list.append(distributable)

    def copyinputs(self, copier):
        for distributable in self._distributable_list:
            copier.input(distributable)

    def copyoutputs(self, copier):
        for distributable in self._distributable_list:
            copier.output(distributable)

     #start of IDistributable interface--------------------------------------
    @property
    def work_count(self):
        return len(self._distributable_list)

    def work_sequence(self):
        for distributable in self._distributable_list:
                yield lambda distributable=distributable : self.do_work(distributable)

    def do_work(self, distributable):
        result = self.indirect_runner.run(distributable)
        return result

    def reduce(self, result_list):
        return list(result_list)

    @property
    def tempdirectory(self):
        if len(self._distributable_list) == 0 : raise Exception("Expect at least one work item when tempdirectory is requested")
        return self._distributable_list[0].tempdirectory

    def __str__(self):
        str_list = []
        for distributable in self._distributable_list:
            str_list.append(str(distributable))
        return "IndirectDistributable({0})".format(",".join(str_list))

     #end of IDistributable interface---------------------------------------

    def __repr__(self):
        import cStringIO
        fp = cStringIO.StringIO()
        fp.write("{0}(\n".format(self.__class__.__name__))
        varlist = []
        for f in dir(self):
            if f.startswith("_"): # remove items that start with '_'
                continue
            if type(self.__class__.__dict__.get(f,None)) is property: # remove @properties
                continue
            if callable(getattr(self, f)): # remove methods
                continue
            varlist.append(f)
        for var in varlist[:-1]: #all but last
            fp.write("\t{0}={1},\n".format(var, getattr(self, var).__repr__()))
        var = varlist[-1] # last
        fp.write("\t{0}={1})\n".format(var, getattr(self, var).__repr__()))
        result = fp.getvalue()
        fp.close()
        return result




'''
Runs a distributable job on multiple processors.  Returns the value of the job.

See SamplePi.py for examples.
'''

from fastlmm.util.runner import *
import os
import logging
try:
    import dill as pickle
except:
    logging.warning("Can't import dill, so won't be able to clusterize lambda expressions. If you try, you'll get this error 'Can't pickle <type 'function'>: attribute lookup __builtin__.function failed'")
    import cPickle as pickle
import subprocess, sys, os.path
import multiprocessing
import fastlmm.util.util as util


class LocalMultiProc: # implements IRunner

    def __init__(self, taskcount, mkl_num_threads = None, just_one_process = False, logging_handler=logging.StreamHandler(sys.stdout)):
        self.just_one_process = just_one_process
        logger = logging.getLogger()
        if not logger.handlers:
            logger.setLevel(logging.INFO)
        for h in list(logger.handlers):
            logger.removeHandler(h)
        if logger.level == logging.NOTSET:
            logger.setLevel(logging.INFO)
        logger.addHandler(logging_handler)

        self.taskcount = taskcount
        self.mkl_num_threads = mkl_num_threads

    def run(self, distributable):
        JustCheckExists().input(distributable)

        localpath = os.environ["PATH"]
        localwd = os.getcwd()

        import datetime
        now = datetime.datetime.now()
        run_dir_rel = os.path.join("runs",util.datestamp(appendrandom=True))
        run_dir_abs = os.path.join(localwd,run_dir_rel)
        util.create_directory_if_necessary(run_dir_rel, isfile=False)

        distributablep_filename = os.path.join(run_dir_rel, "distributable.p")
        with open(distributablep_filename, mode='wb') as f:
            pickle.dump(distributable, f, pickle.HIGHEST_PROTOCOL)

        distributable_py_file = os.path.join(os.path.dirname(__file__),"..","distributable.py")
        if not os.path.exists(distributable_py_file): raise Exception("Expect file at " + distributable_py_file + ", but it doesn't exist.")
        command_format_string = sys.executable + " " + distributable_py_file + " " + distributablep_filename +" LocalInParts({0},{1},mkl_num_threads={2})".format("{0}", self.taskcount, self.mkl_num_threads)

        if not self.just_one_process:
            proc_list = []
            for taskindex in xrange(self.taskcount):
                command_string = command_format_string.format(taskindex)
                proc = subprocess.Popen(command_string.split(" "), cwd=os.getcwd())#!!!bug: If Anaconda is installed in c:\program files\anaconda2 this will fail
                proc_list.append(proc)

            for taskindex, proc in enumerate(proc_list):            
                rc = proc.wait()
                #for line in proc.stdout.readlines():
                #    sys.stdout.write(line)
                if not 0 == rc : raise Exception("Running python in python results in non-zero return code in task#{0}".format(taskindex))
        else:
            from fastlmm.util.runner import LocalInParts
            for taskindex in xrange(self.taskcount):
                LocalInParts(taskindex,self.taskcount, mkl_num_threads=self.mkl_num_threads).run(distributable)

        result = run_one_task(distributable, self.taskcount, self.taskcount, distributable.tempdirectory)


        JustCheckExists().output(distributable)
        return result







exactUpdate = False
logdelta = SP.log(16)
topKbyLinReg = 16

if topKbyLinReg!=None:
    outFile = 'examples/ScanOSP.exact%d.nSnps%d.logdelta%.2f.out.txt'%(exactUpdate,topKbyLinReg,logdelta)
else:
    outFile = 'examples/ScanOSP.exact%d.all.logdelta%.2f.out.txt'%(exactUpdate,logdelta)

obj = ScanOSP(
    phenoFile = 'examples/toydata.phe',
    bedFile = 'examples/toydata',

    windowByPosition = 100,
    logdelta = logdelta,
    topKbyLinReg = topKbyLinReg,
    exactUpdate = exactUpdate,

    #prefilterNumSnps = 1000,
    #extractSim = 'examples/LmmGWAS.snps.txt',

    nFolds = 5,
    nCV = 3,

    outFile = outFile,
    doPlots = True,
    doDebugging = True
)



import fastlmm.feature_selection.PerformSelectionDistributable as psd
import fastlmm.util.preprocess as up
import numpy as np

def set_Gexclude(G_exclude, G1, i_exclude):
        if G_exclude is None and i_exclude is not None and i_exclude.sum()>0:
            assert self.G0 is not None, "i_exclude without SNPs to exclude"
            G = np.hstack((self.G0[:,i_exclude],G1))
            i_G1=np.ones(G.shape[1],dtype='bool')
            n_exclude = i_exclude.sum()
            i_G1[0:n_exclude]=False
        elif G_exclude is not None:
            G = np.hstack((G_exclude,G1))
            i_G1=np.ones(G.shape[1],dtype='bool')
            n_exclude = G_exclude.shape[1]
            i_G1[0:n_exclude]=False
        else:
            G = G1
            i_G1=np.ones(G.shape[1],dtype='bool')
            n_exclude = 0
        return G, i_G1, n_exclude

def set_snps0(SNPs0,sample_size,i_exclude=None, forcefullrank=False,blocksize=10000):
    '''
    In full rank case, loads up the SNPs in blocks, and construct the kernel.
    In low rank case, loads up all SNPs in to memory
    '''    
    if SNPs0 is None:
        return None, None
    if SNPs0.has_key("K"):
        K0 = SNPs0["K"]
        G0 = None
    elif SNPs0.has_key("data"):
        K0 = None
        G0 = SNPs0["data"]["snps"]
    else:        
        #full rank
        if len(SNPs0["snp_set"]) > sample_size or forcefullrank:# N = Y.shape[0]                      
            SNPs0["K"] = psd.build_kernel_blocked(snpreader=SNPs0["reader"], snp_idx=SNPs0["snp_set"].to_index, blocksize=blocksize,allowlowrank=forcefullrank)
            K0 = SNPs0["K"]
            G0 = None
        else:
            #low rank            
            K0 = None
            SNPs0["data"] = SNPs0["snp_set"].read()
            SNPs0["data"]["snps"] = up.standardize(SNPs0["data"]["snps"])
            G0 = SNPs0["data"]["snps"]

    #lrt_up should never do exclusion, because set_snps0 should only get called once, in run_once, without exclusion
    #exclude. So this is only for score test and lrt. 
    if i_exclude is not None:
        if K0 is not None:
            #Also note in the full rank case with exclusion, for score, one could in principle use low rank updates to make it faster,
            #when the number of excluded SNPs is small: it wold be cubic in num_excluded * num_inner*num_outer iterations, versus now
            #where it is cubic in N in the outer loop only once
            K_up = psd.build_kernel_blocked(snpreader=SNPs0["reader"], snp_idx=np.array(SNPs0["snp_set"].to_index)[i_exclude], blocksize=blocksize,allowlowrank=forcefullrank)
            K0 = K0 - K_up
        elif G0 is not None:
            G0 = G0[:,~i_exclude]                        
        num_snps = SNPs0["num_snps"] - i_exclude.sum()
    else:
        num_snps = SNPs0["num_snps"]
    #intersect data?
        
    #normalize:
    if K0 is not None:
        K0 = K0 / num_snps#K0.diagonal().mean()
    elif G0 is not None:
        G0 = G0 / np.sqrt( num_snps )#(G0*G0).mean() ) # computes the sqrt of the mean of the diagonal of K=GG^T; *  means pointwise multiplication 
    return G0, K0


import fastlmm.association.lrt as lr
import scipy as SP
import fastlmm.util.stats.chi2mixture as c2
import fastlmm.association.testCV as testCV
import logging

class Cv(object):
    """description of class"""

    __slots__ = []

    def check_nperm(self,nperm):
        return nperm #cv handles nperm>0 so just return whatever value gomes in

    def __init__(self):
        pass

    def __str__(self):
        #return "cv_{0}".format(self.penalty) horta20130609
        return "cv"

    def construct(self, Y, X=None, forcefullrank = None, SNPs0 = None, nullModel = None, altModel = None,
                  scoring = None, greater_is_better = None):
        return testCV.testCV(Y=Y[:,SP.newaxis],X=X,G0=G0,
                             nullModel=nullModel,altModel=altModel,
                             scoring=scoring, greater_is_better=greater_is_better)

    def construct_no_backgound_kernel(self, Y,X, forcefullrank, nullModel, altModel, scoring,
                                      greater_is_better):
        return construct(Y,X,forcefullrank,None,nullModel,altModel,scoring,greater_is_better)

    def pv(squaredform,expectationsqform,varsqform,GPG):
        raise Exception("'pv' doesn't apply to cv only to davies")

    @property
    def npvals(self):
        return 1 #returns 1 type of p-value

    def w2(self, G0, result):
        return SP.nan

    def lrt(self, result):
        return result.test['stat'] #iset is the index of the test (irrespective of permutation)

    def pv_adj_from_result(self, result):
        return SP.nan;

    def write(self, fp,ind, result_dict, pv_adj, detailed_table):
        fp.write("\t".join(("SetId", "stat","scoreAlt", "scoreNull", "P-value_adjusted", "#SNPs_in_Set", "#ExcludedSNPs", "Chrm", "Pos. range")) + "\n")
        for i in xrange(len(ind)):
            ii = ind[i]
            result = result_dict[ii]
            lik1=result.test['scores'].mean()
            lik0=result.test['scores0'].mean()
            fp.write("\t".join((result.setname, str(result.test['stat']), str(lik1), str(lik0),
                                str(pv_adj[ii]), str(result.setsize),
                                str(result.nexclude), result.ichrm, result.iposrange)
                                + "\n"))


    def pv_etc(self, filenull, G0_to_use, G1, y, x, null_model, varcomp_test, forcefullrank):
        return [SP.nan,SP.nan,SP.nan]

    #!! could these three methods be move to __init__.py?
    @staticmethod
    def pv_adj_and_ind(nperm, pv_adj, nullfit, lrt, lrtperm,
                       alteqnull, alteqnullperm, qmax, nullfitfile,nlocalperm,sort=True):
        
        if nullfit=="ml":
            logging.info("using ml fit for parameterized null distrib")
            assert nullfitfile is None, "not implemented: nullfitfile with ml fit"
            pv_adj,mixture,scale,dof=Cv.lrtpvals_mlfit(nperm, lrt, lrtperm)
        elif nullfit=="qq":
            logging.info("using qq fit for parameterized null distrib")
            pv_adj,mixture,scale,dof=Cv.lrtpvals_qqfit(nperm, lrt, lrtperm,
                                                       alteqnull, alteqnullperm,qmax=qmax,nullfitfile=nullfitfile)
        elif nullfit=="abs":
            assert nullfitfile is None, "not implemented: nullfitfile with abs fit"
            assert nperm==0, "currently can only use abs fit with no permutations"
            logging.info("using qq fit for parameterized null distrib, without using any permutations")
            pv_adj,mixture,scale,dof=Cv.lrtpvals_qqfit(nperm, lrt, lrtperm,
                                                       alteqnull, alteqnullperm,
                                                       abserr=True,fitdof=False,dof=1,qmax=qmax)
        else:
            raise Exception("dont' know nullfit='{0}'".format(nullfit))
        logging.info("mixture (non-zero dof)="+str(mixture))
        logging.info("dof="+str(dof))
        logging.info("scale="+str(scale))
        if sort:
            ind = pv_adj.argsort()
        else:
            ind = None
        return pv_adj, ind

  
    @staticmethod
    def lrtpvals_qqfit(nperm, lrt, lrtperm, alteqnull, alteqnullperm, qmax=None, abserr=False,fitdof=True, dof=None, nullfitfile=None):
        '''
        Fit the parameters of the null distribution using "quantile regession" on some fraction of the most signficant data points
        '''
        if (nperm > 0):            
            logging.info("estimating mixture parameters for permuted data\nusing quantile regression of log-pvalues with qmax=" + str(qmax) + " from permutations...")
            mix = c2.chi2mixture( lrt = lrtperm, qmax = qmax, alteqnull = alteqnullperm,abserr=abserr,fitdof=fitdof,dof=dof)
            
                        
            res = mix.fit_params_Qreg() # paramter fitting             
         
            imax=res['imax']
            mse=res['mse']
            logging.info("# of pvals used for nullfit=" + str(imax))
            pv_adj = mix.sf(lrt=lrt,alteqnull=alteqnull) # getting p-values for real data
            logging.info(" Done")
            logging.info("adjusting the observed p-values ...")

            logging.info(" Done")
        elif nullfitfile is not None:
            logging.info("estimating mixture parameters for permuted data\nusing quantile regression of log-pvalues with qmax=" 
                      + str(qmax) + " from STORED permutations in " + nullfitfile)           
                        
            #read in p-vals and alteqnull for from file                        
            colnames={"2*(LL(alt)-LL(null))","alteqnull","setsize"}
            import fastlmm.util.util as ut
            import numpy as np
            dat=ut.extractcols(nullfitfile,colnameset=colnames,dtypeset={"2*(LL(alt)-LL(null))": np.float64})
            lrtfile=dat["2*(LL(alt)-LL(null))"]
            alteqnullfile=dat["alteqnull"]
            
            mix = c2.chi2mixture( lrt = lrtfile, qmax=qmax, alteqnull=alteqnullfile,abserr=abserr,fitdof=fitdof,dof=dof)
                                    
            res = mix.fit_params_Qreg() # paramter fitting             
         
            imax=res['imax']
            mse=res['mse']
            logging.info("# of pvals used for nullfit=" + str(imax))          
            pv_adj = mix.sf(lrt=lrt,alteqnull=alteqnull) # getting p-values for real data
            logging.info(" Done")
            logging.info("adjusting the observed p-values ...")

            logging.info(" Done")
        else:
            logging.info("nperm = " + str(nperm) + " : No permutations were performed.")
            logging.info("estimating mixture parameters for non-permuted data\nusing quantile regression of log-pvalues with qmax=" + str(qmax) + "...")
            mix = c2.chi2mixture( lrt = lrt, qmax = qmax, alteqnull = alteqnull,abserr=abserr,fitdof=fitdof,dof=dof)
            res = mix.fit_params_Qreg() # paramter fitting
            imax=res['imax']
            logging.info("# of pvals used for nullfit=" + str(imax))
            pv_adj = mix.sf() #getting p-values
            logging.info(" Done")
        if mix.mixture==0:
            #raise Exception("only zero dof component items found")
            logging.info("*****WARNING*****: only zero dof component items found")
        return pv_adj,mix.mixture,mix.scale,mix.dof#,imax,mse

    @staticmethod
    def lrtpvals_mlfit(nperm, a2, a2perm, lrt, lrtperm):
        '''
        Fit the parameters of the null distribution using maximum likelihood on all the data points
        '''
        raise Exception("made changes to use alteqnull and did not modify this code as it looks obsolete")
        if (nperm > 0):
            logging.info("estimating mixture parameters for permuted data using maximum likelihood...")
            # estimating mixture parameters for permuted data, using
            [pvperm,mixture,scale,dof,i0] = c2.computePVmixtureChi2_old(lrtperm,a2perm, tol= 0.0, mixture=None, scale = None, dof = None)
            logging.info("        Done")
            logging.info("adjusting the observed p-values ...")
            # adjust the observed p-values with estimated null mixture, scale, dof parameters
            [pv_adj,mixture,scale,dof,i0] = c2.computePVmixtureChi2_old(lrt,a2, tol= 0.0, mixture=mixture, scale = scale, dof = dof)
            logging.info("        Done")
        else:
            logging.info("nperm = " + str(nperm) + " : No permutations were performed...")
            logging.info("adjusting the observed p-values ...")
            # adjust the observed p-values with estimated null mixture, scale, dof parameters
            [pv_adj,mixture,scale,dof,i0] = c2.computePVmixtureChi2_old(lrt,a2, tol= 0.0, mixture=None, scale = None, dof = None)
            logging.info("        Done")
        return pv_adj,mixture,scale,dof



class Subset(object): # implements ISnpSetList
    '''
    Returns a subset of the originally specified sets to test.
    See the Bed class's 'read' method of examples of its use.
    See __init__.py for specification of interface it implements.
    '''
    def __init__(self, altset_list, subset_list):
        self.subset_list = subset_list
        if isinstance(altset_list, str):#if given a filename, then assumes group-SNP format (default)
            self.inner = SnpAndSetNameCollection(altset_list)
        else:                           #given a NucRangeList(filenam.txt), or any other reader
            self.inner = altset_list

    def addbed(self, bed):
        return SubsetPlusBed(self,bed)

    #would be nicer if these used generic pretty printer
    def __repr__(self):
        return "Subset(altset_list={0},subset_list={1})".format(self.inner,self.subset_list)




class SubsetPlusBed(object): # implements ISnpSetListPlusBed
    '''
    Returns a subset of the originally specified sets to test.
    '''
    def __init__(self, spec, bed):
        self.spec = spec
        self.bed = bed
        self.inner = spec.inner.addbed(bed)

    def __len__(self):
        return len(self.spec.subset_list)

    def __iter__(self):                 #after grabbing main list (self.inner_altset_list), do subsetting
        seenit=[]
        for altset in self.inner:
            if str(altset) in self.spec.subset_list:
               seenit.append(str(altset))
               yield altset
        if not len(seenit)==len(self) : raise Exception("some sets specified not present")



import sys
import pdb
import os
import gzip
import re
import scipy as SP
writesmall = 0

datadir = ""
base = ""
tbi = base+".tbi"
basef=os.path.join(datadir,base)
basefsmall = "chr22.small.txt.gz"
basefsmall = 'chr22.small.txt'
baseftbi = basef + ".tbi"

if writesmall:
    file = gzip.GzipFile(basef, 'r')
    #outfsmall = gzip.GzipFile(basefsmall,'w')
    outfsmall = open(basefsmall,'w')
else:
    file = open(basefsmall,'r')
line = file.readline()
if writesmall:
    outfsmall.write(line)
if line != '##fileformat=VCFv4.1\n':
    print 'wrong format'
iline = 0
header = ''
while (line[0:2]=='##'):
    header = header+line
    line = file.readline()
    if writesmall:
        outfsmall.write(line)
    iline+=1;
if line[0:6] == '#CHROM':
    fields = line[1::].strip().split('\t')
    indID=fields[9::]
    fields=fields[0:9]


famfile = open('test.fam','w')
for id in indID:
   outline = '0 ' + id + ' 0 0 0 -9\n'
   famfile.write(outline)
famfile.close()
ninds = len(indID)

line = file.readline()
if writesmall:
    outfsmall.write(line)

bedfile = open('test.bed','wb')
bimfile = open('test.bim','w')

#init bedfile
bedfile.write(b'l\x1b')#bed header
bedfile.write(b'\x01')
#genomap=[]
#geno = []
iline = 0
if writesmall:
    maxline =100
else:
    maxline = 10000000
while (iline < maxline ) and (line !=''):
    vals = line.strip().split('\t')
    if (vals[6]=='PASS'):
        linebim = vals[0] + ' ' + vals[2] + ' 0 ' + vals[1] + ' ' + vals[3] + ' ' + vals[4] + '\n'
        bimfile.write(linebim)
        #genomap.append(vals[0:9])
        #geno.append(vals[9::])

        #write bedfile entries
        nbytes = SP.ceil(ninds/4.0)
        ndone = 0
        for i in xrange(int(SP.ceil(ninds/4.0))):
            num = min(4,ninds-ndone)
            geno = vals[9+i*4:9+(i)*4+num]
            byte = 0
            #pdb.set_trace()
            for j in xrange(num):

                if geno[j][0:3]=='0|0':
                    #byte += 0*(2**(2*j))
                    pass
                elif geno[j][0:3]=='0|1' or geno[j][0:3]=='1|0':
                    byte += 2*(2**(2*j))
                elif geno[j][0:3]=='1|1':
                    byte += 3*(2**(2*j))
                else:
                    #pdb.set_trace()
                    byte += 1*(2**(2*j))
                pass

            bytestr = '%c'%byte
            bedfile.write(bytestr)
            ndone +=num
            pass
    line = file.readline()
    if writesmall:
        outfsmall.write(line)
    iline+=1
file.close()
if writesmall:
    outfsmall.close()
bimfile.close()
bedfile.close()

#filetbi = open(baseftbi,'r')
#filetbigz = gzip.GzipFile(baseftbi,'r')



import scipy as SP
import numpy as NP
import scipy.linalg as LA
import scipy.optimize as opt
import scipy.stats as ST
import scipy.special as SS
from fastlmm.util.mingrid import *
from fastlmm.util.util import *
import time
import pdb

from bin2kernel import Bin2Kernel
from bin2kernel import makeBin2KernelAsEstimator
from bin2kernel import Bin2KernelLaplaceLinearN
from bin2kernel import getFastestBin2Kernel
from bin2kernel import Bin2KernelEPLinearN

def rotate(A, rotationMatrix,transposeRot=True):
    '''
    Apply an eigenvalue decomposition rotation matrix (Assummes that the rotation-matrix has orthogonal cols)
        
    takes care of low rank structure in A
        
    result = eig[1].T * A
    '''
    [Nr,k] = rotationMatrix.shape
    [N,M] = A.shape
    if Nr!=N:
        raise Exception("rotation and A are not aligned A.shape =  [%i,%i], rotation.shape = [%i,%i]" % (N,M,Nr,k))
    if k>Nr:
        raise Exception("rotation matrix has more columns than rows =  [%i,%i], rotation.shape = [%i,%i]" % (N,M,Nr,k))
        
    res = rotationMatrix.T.dot(A)

    if k<Nr:
        resLowRank = A - rotationMatrix.dot(res)
        res = [res,resLowRank]
    return res

def rotSymm(A,eig,delta=1.0,gamma=1.0, exponent = -0.5, forceSymm = True):
    '''
    do the rotation with K_0 to get a matrix square root
    '''
    [N,M]=A.shape
    [Nr,k] = eig[1].shape
    if Nr!=N:
        raise Exception("rotation and A are not aligned A.shape =  [%i,%i], rotation.shape = [%i,%i]" % (N,M,Nr,k))
    if symmetric and N!=M:
        raise Exception("A is not symmetric. A.shape =  [%i,%i], rotation.shape = [%i,%i]" % (N,M,Nr,k))
    if k>Nr:
        raise Exception("rotation matrix has more columns than rows =  [%i,%i], rotation.shape = [%i,%i]" % (N,M,Nr,k))

    res = eig[1].T.dot(A)
    deltaPow = myPow(delta,exponent,minVal)
    diag = eig[0]*gamma + delta
    diag = myPow(diag,exponent,minVal=0.0)
    if Nr>k:
        diag-=deltaPow
        res = dotDiag(res,diag)
        res = A-eig[1].dot(res)
    else:
        res = dotDiag(res,diag)
        if forceSymm:
            res = eig[1].dot(res)
    return res

def dotDiag(A, diag, minVal = None, symmetric = False, exponent = 1.0):
    '''
    Multiply by a diagonal matrix build from
        
    result = diag.dot(A)
    ''' 
    [N,M] = A.shape
    [Nd] = diag.shape
    if symmetric and N!=M:
        raise Exception("A is not symmetric. A.shape =  [%i,%i], rotation.shape = [%i,%i]" % (N,M,Nr,k))
    if Nd!=N:
        raise Exception("Matrices misaligned A.shape =  [%i,%i], diag.shape = [%i,%i]" % (N,M,Nr,k))
        
    diagcorr = myPow(diag,exponent,minVal)

    result = A * NP.lib.stride_tricks.as_strided(diagcorr, (diagcorr.size,A.shape[1]), (diagcorr.itemsize,0))
    if symmetric:
        result = dotDiag(A=A.T,diag=diagcorr,checkPos=False,minVal = minVal,symmetric = False, exponent = 1.0).T
    return result

def myPow(A,exponent,minVal=None):
    '''
    compute a power, with efficient computation of common powers
    '''
    if exponent == 1.0:
        Aexp = A
    elif exponent == -1.0:
        Aexp = 1.0/A
    elif exponent == 2.0:
        Aexp = A*A
    elif exponent == 0.5:
        Aexp = NP.sqrt(A)
    elif exponent == -0.5:
        Aexp = 1.0/NP.sqrt(A)
    elif exponent == -2.0:
        Aexp = 1.0/(A*A)
    else:
        Aexp = NP.power(A,exponent)
    if minVal is not None:
        i_Null = A<=minVal
        Aexp[i_Null]=0.0
    return Aexp

class lmm2k(object):
    '''
    linear mixed model with up to two kernels
    N(y | X*beta ; sigma2(gamma0*K0 + gamma1*K1 + I ) ),
    where
    K0 = G0*G0^T
    K1 = G1*G1^T
    '''
    __slots__ = ["G0","G1","Y","X","K0","K1","K","U","S","UX","Uy","UUX","UW","UUW","UUy","pos0","pos1","gamma0","gamma1","delta","exclude_idx","forcefullrank","numcalls","Xstar","Kstar","Kstar_star","UKstar","UUKstar","Gstar"]

    def __init__(self,forcefullrank=False):
        '''
        Input:
        forcefullrank   : if True, then the code always computes K and runs cubically
                            (False)
        '''
        self.X=None
        self.Y=None
        self.G0=None
        self.G1=None
        self.K0=None
        self.K1=None
        self.gamma0=1.0
        self.gamma1=1.0
        self.delta=1.0
        self.eig0=None
        self.eig1=None
        self.Yrot = None
        self.Xrot = None
        self.K1rot = None
        self.G1rot = None

    def setX(self, X):
        '''
        set the fixed effects X (covariates).
        The Kernel has to be set in advance by first calling setG() or setK().
        --------------------------------------------------------------------------
        Input:
        X       : [N*D] 2-dimensional array of covariates
        --------------------------------------------------------------------------
        '''
        self.X    = X
        self.Xrot = None
        
    def setY(self, Y):
        '''
        set the phenotype y.
        The Kernel has to be set in advance by first calling setG() or setK().
        --------------------------------------------------------------------------
        Input:
        Y       : [NxP] P-dimensional array of phenotype values
        --------------------------------------------------------------------------
        '''
        self.Y    = Y
        self.Yrot = None
        
    def setG0(self, G0):
        '''
        set the Kernel K0 from G0.
        This has to be done before setting the data setX() and setY(). 
        ----------------------------------------------------------------------------
        Input:
        G0              : [N*k0] array of random effects
        -----------------------------------------------------------------------------
        '''
        k = G0.shape[1]
        N = G0.shape[0]
        self.eig0 = None
        self.eig1 = None
        self.K1rot = None
        self.Yrot = None
        self.Xrot = None
        
        if ((not self.forcefullrank) and (k<N)):
            self.G0 = G0
        else:
            K0=G0.dot(G0.T);
            self.setK0(K0=K0)
        
    def setK0(self, K0):
        '''
        set the background Kernel K0.
        --------------------------------------------------------------------------
        Input:
        K0 : [N*N] array, random effects covariance (positive semi-definite)
        --------------------------------------------------------------------------
        '''
        self.K0    = K0
        self.G1    = None
        self.eig0  = None
        self.eig1  = None
        self.K1rot = None
        self.Yrot  = None
        self.Xrot  = None
        self.G1rot = None
        self.K1rot = None

    def setG1(self, G1):
        '''
        set the Kernel K1 from G1.
        ----------------------------------------------------------------------------
        Input:
        G0              : [N*k0] array of random effects
        -----------------------------------------------------------------------------
        '''
        k = self.G1.shape[1]
        N = self.G0.shape[0]
        self.eig1 = None
        self.G1rot = None
        if ((not self.forcefullrank) and (k<N)):
            #it is faster using the eigen decomposition of G.T*G but this is more accurate
            self.G1 = G1
        else:
            K1=self.G1.dot(self.G1.T);
            self.setK1(K1=K1)
        pass

    def setK1(self, K1):
        '''
        set the foreground Kernel K1.
        This has to be done before setting the data setX() and setY().
        --------------------------------------------------------------------------
        Input:
        K1 : [N*N] array, random effects covariance (positive semi-definite)
        --------------------------------------------------------------------------
        '''
        self.eig1 = None
        self.G1   = None
        self.K1   = K1
        self.K1rot= None
        self.G1rot= None
        
    def setVariances(self,gamma0=None,gamma1=None,delta=None):
        if gamma0 is not None:
            #background model changed, foreground rotation (U1,S1) changes:
            self.gamma0 = gamma0
            self.eig1 = None
            self.Xrot = None
            self.Yrot = None
            self.K1rot = None
            self.G1rot = None

        if delta is not None:
            #background model changed, foreground rotation (U1,S1) changes:
            self.delta = delta
            self.eig1 = None
            self.Xrot = None
            self.Yrot = None
            self.K1rot = None
            self.G1rot = None

        if gamma1 is not None:
            #foreground model changed
            self.gamma1 = gamma1




    def getEig1(self):
        '''
        '''
        if self.eig1 is None:
            #compute eig1
            if self.K1 is not None:
                if self.K1rot is None:
                    self.K1rot = rotSymm(self.K1, eig = self.eig0, exponent = -0.5, gamma=self.gamma0,delta = self.delta,forceSymm = False)
                    self.K1rot = rotSymm(self.K1rot.T, eig = self.eig0, exponent = -0.5, gamma=self.gamma0,delta = self.delta,forceSymm = False)
                self.eig1 = LA.eigh(self.K1rot)
            elif self.G1 is not None:
                [N,k] = self.G1.shape
                if self.G1rot is None:
                    self.G1rot = rotSymm(self.G1, eig = self.eig0, exponent = -0.5, gamma=self.gamma0,delta = self.delta,forceSymm = False)
                
                try:
                    [U,S,V] = LA.svd(self.G1rot,full_matrices = False)
                    self.eig1 = [S*S,U]
                except LA.LinAlgError:  # revert to Eigenvalue decomposition
                    print "Got SVD exception, trying eigenvalue decomposition of square of G. Note that this is a little bit less accurate"
                    [S_,V_] = LA.eigh(self.G1rot.T.dot(self.G1rot))
                    S_nonz=(S_>0.0)
                    S1 = S_[S_nonz]
                    U1=self.G1rot.dot(V_[:,S_nonz]/SP.sqrt(S1))
                    self.eig1=[S1,U1]
        return self.eig1

    def getEig0(self):
        '''
        '''
        if self.eig0 is None:
            #compute eig0
            if self.K0 is not None:
                self.eig1 = LA.eigh(self.K0)
            elif self.G1 is not None:
                [N,k] = self.G0.shape
                try:
                    [U,S,V] = LA.svd(self.G0,full_matrices = False)
                    self.eig0 = [S*S,U]
                except LA.LinAlgError:  # revert to Eigenvalue decomposition
                    print "Got SVD exception, trying eigenvalue decomposition of square of G. Note that this is a little bit less accurate"
                    [S_,V_] = LA.eigh(self.G0.T.dot(self.G0))
                    S_nonz=(S_>0.0)
                    S0 = S_[S_nonz]
                    U0=self.G0.dot(V_[:,S_nonz]/SP.sqrt(S0))
                    self.eig0=[S0,U0]
        return self.eig1

    def set_exclude_idx(self, idx):
        '''
        Set the indices of SNPs to be removed
        --------------------------------------------------------------------------
        Input:
        idx  : [k_up: number of SNPs to be removed] holds the indices of SNPs to be removed
        --------------------------------------------------------------------------
        '''        
        self.exclude_idx = idx
        
        
    def findGamma1givenGamma0(self, gamma0 = 0.0, nGridGamma1=10, minGamma1=0.0, maxGamma1=10000.0, **kwargs):
        '''
        Find the optimal h2 for a given K. Note that this is the single kernel case. So there is no a2.
        (default maxH2 value is set to a value smaller than 1 to avoid loss of positive definiteness of the final model covariance)
        --------------------------------------------------------------------------
        Input:
        nGridH2 : number of h2-grid points to evaluate the negative log-likelihood at
        minH2   : minimum value for h2 optmization
        maxH2   : maximum value for h2 optmization
        --------------------------------------------------------------------------
        Output:
        dictionary containing the model parameters at the optimal h2
        --------------------------------------------------------------------------
        '''
        self.setVariances(gamma0 = gamma0)
        resmin=[None]
        def f(x,resmin=resmin,**kwargs):
            #self.setVariances(gamma1=x)
            res = self.nLLeval(gamma1=x,**kwargs)
            if (resmin[0] is None) or (res['nLL']<resmin[0]['nLL']):
                resmin[0]=res
                self.setGamma1[x]
            return res['nLL']
        min = minimize1D(f=f, nGrid=nGridGamma1, minval=minGamma1, maxval=maxGamma1 )
        return resmin[0]

    def findGammas(self, nGridGamma0=10, minGamma0=0.0, maxGamma0=10000.0, nGridGamma1=10, minGamma1=0.0, maxGamma1=10000.0,verbose=False, **kwargs):
        '''
        Find the optimal gamma0 and gamma1, such that K=gamma0*K0+gamma1*K1. Performs a double loop optimization (could be expensive for large grid-sizes)
        --------------------------------------------------------------------------
        Input:
        nGridGamma0 : number of a2-grid points to evaluate the negative log-likelihood at
        minGamma0   : minimum value for a2 optmization
        maxGamma0   : maximum value for a2 optmization
        nGridGamma1 : number of h2-grid points to evaluate the negative log-likelihood at
        minGamma1   : minimum value for h2 optmization
        maxGamma1   : maximum value for h2 optmization
        --------------------------------------------------------------------------
        Output:
        dictionary containing the model parameters at the optimal gamma0 and gamma1
        --------------------------------------------------------------------------
        '''
        self.numcalls=0
        resmin=[None]
        def f(x,resmin=resmin, nGridGamma1=nGridGamma1, minGamma1=minGamma1, maxGamma1=maxGamma1,**kwargs):
            self.numcalls+=1
            t0=time.time()
            res = self.findGamma1givenGamma0(gamma0=x, nGridGamma1=nGridGamma1, minGamma1=minGamma1, maxGamma1=maxGamma1,**kwargs)
            if (resmin[0] is None) or (res['nLL']<resmin[0]['nLL']):
                resmin[0]=res
            t1=time.time()
            #print "one objective function call took %.2f seconds elapsed" % (t1-t0)
            #import pdb; pdb.set_trace()
            return res['nLL']
        if verbose: print "findGammas"
        min = minimize1D(f=f, nGrid=nGridGamma1, minval=minGamma1, maxval=maxGamma1,verbose=False)
        #print "numcalls to innerLoopTwoKernel= " + str(self.numcalls)
        return resmin[0]


    def nLLeval(self,REML=True, gamma1 = None, dof = None, scale = 1.0):
        '''
        evaluate -ln( N( U^T*y | U^T*X*beta , h2*S + (1-h2)*I ) ),
        where ((1-a2)*K0 + a2*K1) = USU^T
        --------------------------------------------------------------------------
        Input:
        h2      : mixture weight between K and Identity (environmental noise)
        REML    : boolean
                  if True   : compute REML
                  if False  : compute ML
        dof     : Degrees of freedom of the Multivariate student-t
                        (default None uses multivariate Normal likelihood)
        logdelta: log(delta) allows to optionally parameterize in delta space
        delta   : delta     allows tomoptionally parameterize in delta space
        scale   : Scale parameter the multiplies the Covariance matrix (default 1.0)
        --------------------------------------------------------------------------
        Output dictionary:
        'nLL'       : negative log-likelihood
        'sigma2'    : the model variance sigma^2
        'beta'      : [D*1] array of fixed effects weights beta
        'h2'        : mixture weight between Covariance and noise
        'REML'      : True: REML was computed, False: ML was computed
        'a2'        : mixture weight between K0 and K1
        'dof'       : Degrees of freedom of the Multivariate student-t
                        (default None uses multivariate Normal likelihood)
        'scale'     : Scale parameter that multiplies the Covariance matrix (default 1.0)
        --------------------------------------------------------------------------
        '''
        if gamma1 is None:
            gamma1=self.gamma1
        if (gamma1<0.0):
            return {'nLL':3E20,
                    'gamma1':gamma1,
                    'delta' : self.delta,
                    'gamma0': self.gamma0,
                    'dof' : dof,
                    'REML':REML,
                    'scale':scale}
        k=self.S1.shape[0]
        N=self.Y.shape[0]
        D=self.X.shape[1]

        Sd = (gamma1*self.S1+1.0)*scale
        UXS = dotDiag(self.UX, Sd, minVal = 0.0, symmetric = False, exponent = -1.0)
        UYS = dotDiag(self.UY, Sd, minVal = 0.0, symmetric = False, exponent = -1.0)

        XKX = UXS.T.dot(self.UX)
        XKy = UXS.T.dot(self.Uy)
        yKy = UyS.T.dot(self.Uy)

        logdetK = SP.log(Sd).sum()


        if (k<N):#low rank part
        
            # determine normalization factor
            denom = (1.0*scale)
            

            XKX += self.UUX.T.dot(self.UUX)/(denom)
            XKy += self.UUX.T.dot(self.UUy)/(denom)
            yKy += self.UUy.T.dot(self.UUy)/(denom)      
            logdetK+=(N-k) * SP.log(denom)
 
        # proximal contamination (see Supplement Note 2: An Efficient Algorithm for Avoiding Proximal Contamination)
        # available at: http://www.nature.com/nmeth/journal/v9/n6/extref/nmeth.2037-S1.pdf
        # exclude SNPs from the RRM in the likelihood evaluation
        

        if len(self.exclude_idx) > 0:
            raise Exception("not implemented")          
            num_exclude = len(self.exclude_idx)
            
            # consider only excluded SNPs
            G_exclude = self.G[:,self.exclude_idx]
            
            self.UW = self.U.T.dot(G_exclude) # needed for proximal contamination
            UWS = self.UW / NP.lib.stride_tricks.as_strided(Sd, (Sd.size,num_exclude), (Sd.itemsize,0))
            assert UWS.shape == (k, num_exclude)
            
            WW = NP.eye(num_exclude) - UWS.T.dot(self.UW)
            WX = UWS.T.dot(self.UX)
            Wy = UWS.T.dot(self.Uy)
            assert WW.shape == (num_exclude, num_exclude)
            assert WX.shape == (num_exclude, D)
            assert Wy.shape == (num_exclude,)
            
            if (k<N):#low rank part
            
                self.UUW = G_exclude - self.U.dot(self.UW)
                
                WW += self.UUW.T.dot(self.UUW)/denom
                WX += self.UUW.T.dot(self.UUX)/denom
                Wy += self.UUW.T.dot(self.UUy)/denom
            
            
            #TODO: do cholesky, if fails do eigh
            # compute inverse efficiently
            [S_WW,U_WW] = LA.eigh(WW)
            
            UWX = U_WW.T.dot(WX)
            UWy = U_WW.T.dot(Wy)
            assert UWX.shape == (num_exclude, D)
            assert UWy.shape == (num_exclude,)
            
            # compute S_WW^{-1} * UWX
            WX = UWX / NP.lib.stride_tricks.as_strided(S_WW, (S_WW.size,UWX.shape[1]), (S_WW.itemsize,0))
            # compute S_WW^{-1} * UWy
            Wy = UWy / S_WW
            # determinant update
            logdetK += SP.log(S_WW).sum()
            assert WX.shape == (num_exclude, D)
            assert Wy.shape == (num_exclude,)
            
            # perform updates (instantiations for a and b in Equation (1.5) of Supplement)
            yKy += UWy.T.dot(Wy)
            XKy += UWX.T.dot(Wy)
            XKX += UWX.T.dot(WX)
            

        #######

        [SxKx,UxKx]= LA.eigh(XKX)
        i_pos = SxKx>1E-10
        beta = SP.dot(UxKx[:,i_pos],(SP.dot(UxKx[:,i_pos].T,XKy)/SxKx[i_pos]))

        r2 = yKy-XKy.dot(beta)

        if dof is None:#Use the Multivariate Gaussian
            if REML:
                XX = self.X.T.dot(self.X)
                [Sxx,Uxx]= LA.eigh(XX)
                logdetXX  = SP.log(Sxx).sum()
                logdetXKX = SP.log(SxKx).sum()
                sigma2 = r2 / (N - D)
                nLL =  0.5 * ( logdetK + logdetXKX - logdetXX + (N-D) * ( SP.log(2.0*SP.pi*sigma2) + 1 ) )
            else:
                sigma2 = r2 / (N)
                nLL =  0.5 * ( logdetK + N * ( SP.log(2.0*SP.pi*sigma2) + 1 ) )
            result = {
                  'nLL':nLL,
                  'sigma2':sigma2,
                  'beta':beta,
                  'gamma1':gamma1,
                  'REML':REML,
                  'gamma0':self.gamma0,
                  'delta':self.delta,
                  'scale':scale
                  }
        else:#Use multivariate student-t
            if REML:
                XX = self.X.T.dot(self.X)
                [Sxx,Uxx]= LA.eigh(XX)
                logdetXX  = SP.log(Sxx).sum()
                logdetXKX = SP.log(SxKx).sum()

                nLL =  0.5 * ( logdetK + logdetXKX - logdetXX + (dof + (N-D)) * SP.log(1.0+r2/dof) )
                nLL += 0.5 * (N-D)*SP.log( dof*SP.pi ) + SS.gammaln( 0.5*dof ) - SS.gammaln( 0.5* (dof + (N-D) ))
            else:
                nLL =   0.5 * ( logdetK + (dof + N) * SP.log(1.0+r2/dof) )
                nLL +=  0.5 * N*SP.log( dof*SP.pi ) + SS.gammaln( 0.5*dof ) - SS.gammaln( 0.5* (dof + N ))
            result = {
                  'nLL':nLL,
                  'dof':dof,
                  'sigma2':sigma2,
                  'beta':beta,
                  'gamma1':gamma1,
                  'REML':REML,
                  'gamma0':self.gamma0,
                  'delta':self.delta,
                  'scale':scale
                  }

        return result




"""
example of how to use feature selection from python (see also command line interface)
"""

import numpy as np
from fastlmm.feature_selection import FeatureSelectionStrategy
from pysnptools.snpreader import Bed

import logging

def runselect(bed_fn=None, pheno_fn=None, strategy=None, select_by_ll=True, output_prefix=None,num_pcs=0, random_state=3, num_snps_in_memory=1000, cov_fn=None, k_values=None, delta_values=None,num_folds=10,penalty=0.0):
    logging.basicConfig(level=logging.INFO)

    # set up data
    ##############################
    if bed_fn is None:
        bed_fn = Bed("examples/toydata")
        pheno_fn = "examples/toydata.phe"
    

    # set up grid
    ##############################
    num_steps_delta = 10
    num_steps_k = 5

    # log_2 space and all SNPs
    #k_values = np.logspace(0, 9, base=2, num=num_steps_k, endpoint=True).tolist() + [10000]
    if k_values is None:
        k_values = [0, 1, 5, 10, 20, 50, 100, 500, 1000, 2000, 5000, 10000, 456345643256] #100, 110, 120, 130, 140, 150, 160, 170, 180, 190, 200, 210, 220, 230, 240, 250, 260, 270, 280, 290, 300, 400, 500, 1000]
    if delta_values is None:
        delta_values = np.logspace(-10, 10, endpoint=True, num=num_steps_delta, base=np.exp(1))
    #delta_values = [np.exp(1), np.exp(2), np.exp(3), np.exp(4), np.exp(5), np.exp(6)]

    if strategy is None:
        strategy = 'lmm_full_cv'
        select_by_ll = True
    if 0:
        strategy = 'insample_cv'
        select_by_ll = True
    # where to save output
    ##############################
    if output_prefix is None:
        output_prefix = "example_pc%i" % (num_pcs)
    
    # go!
    fss = FeatureSelectionStrategy(bed_fn, pheno_fn, num_folds, random_state=random_state, num_pcs=num_pcs, num_snps_in_memory=num_snps_in_memory, interpolate_delta=False, cov_fn=cov_fn)

    best_k, best_delta, best_obj, best_snps = fss.perform_selection(k_values, delta_values, output_prefix=output_prefix, select_by_ll=select_by_ll, strategy = strategy, penalty=penalty)
    res = {
           'best_k':best_k,
           'best_delta':best_delta,
           'best_obj':best_obj, 
           'best_snps':best_snps
           }
    return res
    

if __name__ == "__main__":
    result = runselect()




def qf(chi2val, coeffs,dof = None,noncentrality = None,sigma = 0.0,lim=1000000,acc=1e-08):
    import numpy as np
    from fastlmm.util.stats.quadform.qfc_src import wrap_qfc
    size = coeffs.shape[0]
    if dof is None:
        dof = np.ones(size,dtype = 'int32')
        #dof = np.ones(size)
    if noncentrality is None:
        noncentrality = np.zeros(size)
    ifault=np.zeros(1,dtype = 'int32')
    #ifault=np.zeros(1)
    trace = np.zeros(7)
    #import pdb
    #pdb.set_trace()
    pval = 1.0-wrap_qfc.qf(coeffs,noncentrality,dof,sigma,chi2val,lim,acc,trace,ifault)
    return pval, ifault[0], trace
    
if __name__ == "__main__":
    mix = np.ones(1)
    c = 1.0
    res = qf(c,mix)



import numpy as np
import logging
from fastlmm.association import single_snp
from sklearn.model_selection import KFold
import pandas as pd
import os
import time

import pysnptools.util as pstutil
from fastlmm.inference import FastLMM
from fastlmm.util.mapreduce import map_reduce
from fastlmm.inference.fastlmm_predictor import _snps_fixup, _pheno_fixup, _kernel_fixup
from fastlmm.association.single_snp import _K_per_chrom
from pysnptools.standardizer import Unit

from pysnptools.kernelreader import KernelReader
from pysnptools.kernelreader import KernelData
from fastlmm.association import single_snp_linreg

from fastlmm.util.mapreduce import map_reduce
from fastlmm.association.single_snp_all_plus_select import _kfold
from fastlmm.util.runner import Local
from pysnptools.kernelreader import Identity as KernelIdentity


def _fixup(test_snps, G, pheno, covar,count_A1=None):
    test_snps = _snps_fixup(test_snps,count_A1=count_A1)
    G = _snps_fixup(G or test_snps,count_A1=count_A1)
    pheno = _pheno_fixup(pheno,count_A1=count_A1).read()
    assert pheno.sid_count == 1, "Expect pheno to be just one variable"
    pheno = pheno[(pheno.val==pheno.val)[:,0],:]
    covar = _pheno_fixup(covar, iid_if_none=pheno.iid,count_A1=count_A1)
    G, test_snps, pheno, covar  = pstutil.intersect_apply([G, test_snps, pheno, covar])
    return test_snps, G, pheno, covar

def single_snp_select(test_snps, pheno, G=None, covar=None,
                 k_list = None,
                 n_folds=10, #1 is special and means test on train
                 just_return_selected_snps=False,
                 seed = 0, output_file_name = None,
                 GB_goal=None, force_full_rank=False, force_low_rank=False, h2=None, runner=None, count_A1=None):
    """
    Function performing single SNP GWAS based on covariates (often PCs) and a similarity matrix constructed of the top *k* SNPs where
    SNPs are ordered via the PValue from :meth:`.single_snp_linreg` and *k* is determined via out-of-sample prediction. Will reorder and intersect IIDs as needed.

    :param test_snps: SNPs to test. Can be any :class:`.SnpReader`. If you give a string, it should be the base name of a set of PLINK Bed-formatted files.
           (For backwards compatibility can also be dictionary with keys 'vals', 'iid', 'header')
    :type test_snps: a :class:`.SnpReader` or a string

    :param pheno: A single phenotype: Can be any :class:`.SnpReader`, for example, :class:`.Pheno` or :class:`.SnpData`.
           If you give a string, it should be the file name of a PLINK phenotype-formatted file.
           Any IIDs with missing values will be removed.
           (For backwards compatibility can also be dictionary with keys 'vals', 'iid', 'header')
    :type pheno: a :class:`.SnpReader` or a string

    :param G: SNPs from which to create a similarity matrix of the top *k* SNPs. If not given, will use test_snps.
           Can be any :class:`.SnpReader`. If you give a string, it should be the base name of a set of PLINK Bed-formatted files.
    :type G: :class:`.SnpReader` or a string

    :param covar: covariate information, optional: Can be any :class:`.SnpReader`, for example, :class:`.Pheno` or :class:`.SnpData`.
           If you give a string, it should be the file name of a PLINK phenotype-formatted file.
           (For backwards compatibility can also be dictionary with keys 'vals', 'iid', 'header')
    :type covar: a :class:`.SnpReader` or a string

    :param k_list: Values of *k* (in addition to 0) to test. Default to [1,2,4,8,...8192].
    :type k_list: list of numbers

    :param n_folds: Number of folds of cross validation to use for out-of-sample evaluation of various values of *k*. Default to 10.
    :type n_folds: number
    
    :param just_return_selected_snps: Instead of returning the results of GWAS, return the top *k* SNPs selected.
    :type just_return_selected_snps: list of strings

    :param seed: (optional) Random seed used to generate permutations for lrt G0 fitting.
    :type seed: number

    :param output_file_name: Name of file to write results to, optional. If not given, no output file will be created.
    :type output_file_name: file name

    :param GB_goal: gigabytes of memory the run should use, optional. If not given, will read the test_snps in blocks the same size as the kernel,
        which is memory efficient with little overhead on computation time.
    :type GB_goal: number

    :param force_full_rank: Even if kernels are defined with fewer SNPs than IIDs, create an explicit iid_count x iid_count kernel. Cannot be True if force_low_rank is True.
    :type force_full_rank: Boolean

    :param force_low_rank: Even if kernels are defined with fewer IIDs than SNPs, create a low-rank iid_count x sid_count kernel. Cannot be True if force_full_rank is True.
    :type force_low_rank: Boolean

    :param h2: A parameter to LMM learning that tells how much weight to give the K's vs. the identity matrix, optional
            If not given will search for best value.
    :type h2: number

    :param runner: a runner, optional: Tells how to run locally, multi-processor, or on a cluster.
        If not given, the function is run locally.
    :type runner: a runner.

    :param count_A1: If it needs to read SNP data from a BED-formatted file, tells if it should count the number of A1
         alleles (the PLINK standard) or the number of A2 alleles. False is the current default, but in the future the default will change to True.
    :type count_A1: bool


    :rtype: Pandas dataframe with one row per test SNP. Columns include "PValue"


    :Example:

    >>> import logging
    >>> import numpy as np
    >>> from fastlmm.association import single_snp_select
    >>> from pysnptools.snpreader import Bed
    >>> from fastlmm.util import compute_auto_pcs
    >>> bed_fn = "../../tests/datasets/synth/all.bed"
    >>> phen_fn = "../../tests/datasets/synth/pheno_10_causals.txt"
    >>> covar = compute_auto_pcs(bed_fn,count_A1=False)
    >>> results_dataframe = single_snp_select(test_snps=bed_fn, G=bed_fn, pheno=phen_fn, covar=covar, GB_goal=2, count_A1=False)
    >>> print results_dataframe.iloc[0].SNP,round(results_dataframe.iloc[0].PValue,7),len(results_dataframe)
    snp495_m0_.01m1_.04 0.0 5000

    """

    #!!!code similar to single_snp and feature_selection
    if force_full_rank and force_low_rank:
        raise Exception("Can't force both full rank and low rank")

    assert test_snps is not None, "test_snps must be given as input"

    if k_list is None:
        k_list = np.logspace(start=0, stop=13, num=14, base=2)

    test_snps, G, pheno, covar = _fixup(test_snps, G, pheno, covar, count_A1=count_A1)
    common_input_files = [test_snps, G, pheno, covar]

    k_list_in = [0] + [int(k) for k in k_list if 0 < k and k <= G.sid_count]

    def top_snps_for_each_fold_nested(kfold_item):
        fold_index, (train_idx, test_idx) = kfold_item
        _, G_in, pheno_in, covar_in = _fixup(test_snps, G, pheno, covar,count_A1=count_A1)
        nested = single_snp_linreg(G_in[train_idx,:],pheno_in[train_idx,:],covar_in[train_idx,:],GB_goal=GB_goal,max_output_len=max(k_list_in),count_A1=count_A1)
        return nested
    def top_snps_for_each_fold_reducer(dataframe_list):
        result = [list(dataframe['SNP']) for dataframe in dataframe_list]
        return result


    #Find top snps for each fold        
    fold_index_to_top_snps = map_reduce(_kfold(G.iid_count, n_folds, seed, end_with_all=True,iid_to_index=G.iid_to_index),
                                         nested=top_snps_for_each_fold_nested,
                                         reducer=top_snps_for_each_fold_reducer,
                                         name = "top_snps_for_each_fold",
                                         input_files = common_input_files,
                                         runner=runner)

    #=================================================
    # Start of definition of inner functions
    #=================================================
    def k_index_to_nLL_mapper(k):
        _, G_in, pheno_in, covar_in = _fixup(test_snps, G, pheno, covar,count_A1=count_A1)
        nll_sum=0
        mse_sum = 0
        n_folds_in = 0
        for fold_index, (train_idx, test_idx) in _kfold(G.iid_count, n_folds, seed, end_with_all=False,iid_to_index=G.iid_to_index):
            assert set(train_idx).isdisjoint(set(test_idx)), "real assert"
            top_snps_in_fold = fold_index_to_top_snps[fold_index][:k]
            sid_idx_in_fold = G_in.sid_to_index(top_snps_in_fold)
            G_train = G_in[train_idx,sid_idx_in_fold] if k > 0 else None
            fastlmm = FastLMM(force_full_rank=force_full_rank, force_low_rank=force_low_rank,GB_goal=GB_goal)
            fastlmm.fit(K0_train=G_train, X=covar_in[train_idx,:], y=pheno_in[train_idx,:], h2raw=h2) #iid intersection means when can give the whole covariate and pheno
            G_test = G_in[test_idx,sid_idx_in_fold] if k > 0 else KernelIdentity(G_in.iid,G_in.iid[test_idx]) #!!! instead of this, which blows up when # of iids is large, should switch to linear regression model with k is 0
            nll,mse = fastlmm.score(K0_whole_test=G_test,X=covar_in[test_idx,:],y=pheno_in[test_idx,:],return_mse_too=True) #iid intersection means when can give the whole covariate and pheno
            nll_sum += nll
            mse_sum += mse
            n_folds_in += 1
        logging.info("k={0},nLL={1},average mse={2}".format(k,nll_sum,mse_sum / n_folds_in))
        return nll_sum
    #=================================================
    # End of definition of inner functions
    #=================================================


    #find best # of top SNPs
    k_index_to_nLL = map_reduce(k_list_in,
                                mapper = k_index_to_nLL_mapper,
                                input_files = common_input_files,
                                name = "k_index_to_nLL",
                                runner=runner)
    best_k = k_list_in[np.argmin(k_index_to_nLL)]

    top_snps = fold_index_to_top_snps[-1][:best_k]
    if just_return_selected_snps:
        return top_snps

    sid_idx = G.sid_to_index(top_snps)
    G_top = G[:,sid_idx]

    # Run GWAS with leave-one-chrom out
    single_snp_result = single_snp(test_snps=test_snps, K0=G_top, pheno=pheno,
                                covar=covar, leave_out_one_chrom=True,
                                GB_goal=GB_goal,  force_full_rank=force_full_rank, force_low_rank=force_low_rank, h2=h2,
                                output_file_name=output_file_name,runner=runner,count_A1=count_A1)

    return single_snp_result

if __name__ == "__main__":

    logging.basicConfig(level=logging.WARN)

    import doctest
    doctest.testmod()

